"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import abc
import asyncio
import contextlib
import datetime
import enum
import itertools
import logging
import typing
import uuid
from abc import abstractmethod
from collections.abc import Collection, Mapping, Sequence, Set
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Optional, Self

import asyncpg

from inmanta import const, data
from inmanta.agent import executor
from inmanta.agent.code_manager import CodeManager
from inmanta.data import ConfigurationModel, Environment
from inmanta.data.model import Discrepancy, SchedulerStatusReport
from inmanta.deploy import timers, work
from inmanta.deploy.persistence import ToDbUpdateManager
from inmanta.deploy.state import AgentStatus, Blocked, Compliance, DeployResult, ModelState, ResourceIntent, ResourceState
from inmanta.deploy.tasks import Deploy, DryRun, RefreshFact, Task
from inmanta.deploy.work import TaskPriority
from inmanta.protocol import Client
from inmanta.resources import Id
from inmanta.types import ResourceIdStr, ResourceType, ResourceVersionIdStr

LOGGER = logging.getLogger(__name__)
NB_ITERATIONS_PASS_IO_LOOP: int = 100


class StaleResource(Exception):
    """
    An exception that indicates that a resource is not managed by the orchestrator.
    """

    pass


@dataclass(frozen=True)
class ResourceVersionIntent:
    """
    Resource intent for a single resource at a specific version.

    Includes model version and resource intent, as well as the full set of resource types that live on the resource's agent.
    This list will never be empty, and will always include the resource's own type.
    """

    model_version: int
    intent: ResourceIntent
    # All types that live on this agent. Required to ensure that the executor loads the appropriate code for this version,
    # even if new versions come in before the executor is constructed.
    # Should be dropped once this functionality moves to the code manager.
    # At that point, the state's types_per_agent can be dropped as well.
    all_types_for_agent: Collection[ResourceType]


@dataclass(frozen=True)
class DeployIntent(ResourceVersionIntent):
    """
    Deploy intent for a single resource. Includes dependency state to provide to the resource handler.
    """

    dependencies: Mapping[ResourceIdStr, const.ResourceState]
    deploy_start: datetime.datetime


class ResourceIntentChangeType(Enum):
    """
    A state change for a single resource's intent. Represents in which way, if any, a resource changed in a new model version
    versus the currently managed one.

    Internal class meant to unambiguously categorize how resources have changed when processing multiple model versions at once.
    """

    NEW = enum.auto()
    """
    To be considered a new resource, even if one with the same resource id is already managed in the current version.
    e.g. if multiple versions have been released since the currently managed one, and in those a resource is deleted,
    then reappears, we consider it new for all purposes.
    """

    UPDATED = enum.auto()
    """
    The resource has an update to its desired state, this may include a transition from defined to undefined or vice versa.
    """

    DELETED = enum.auto()
    """
    The resource was deleted.
    """


class _ResourceIntentChange(abc.ABC):
    change: ClassVar[ResourceIntentChangeType]


class New(_ResourceIntentChange):
    change = ResourceIntentChangeType.NEW


class Updated(_ResourceIntentChange):
    change = ResourceIntentChangeType.UPDATED


@dataclass(frozen=True, kw_only=True)
class Deleted(_ResourceIntentChange):
    change: ClassVar[ResourceIntentChangeType] = ResourceIntentChangeType.DELETED

    last_seen_version: int


type ResourceIntentChange = New | Updated | Deleted


class ResourceRecord(typing.TypedDict):
    """
    A dict representing a resource database record with all fields relevant for the scheduler.

    Purely for type documentation purposes, as we can't statically verify it, considering asyncpg's limitations.
    """

    resource_id: str
    status: str
    attributes: Mapping[str, object]
    attribute_hash: str


@dataclass(frozen=True)
class ModelVersion:
    """
    A version of the model to be managed by the scheduler.
    """

    version: int
    resources: Mapping[ResourceIdStr, ResourceIntent]
    requires: Mapping[ResourceIdStr, Set[ResourceIdStr]]
    undefined: Set[ResourceIdStr]

    @classmethod
    def from_db_records(cls: type[Self], version: int, resources: Collection[ResourceRecord]) -> Self:
        return cls(
            version=version,
            resources={
                ResourceIdStr(resource["resource_id"]): ResourceIntent(
                    resource_id=ResourceIdStr(resource["resource_id"]),
                    attribute_hash=resource["attribute_hash"],
                    attributes=resource["attributes"],
                )
                for resource in resources
            },
            requires={
                ResourceIdStr(resource["resource_id"]): {
                    Id.parse_id(req).resource_str() for req in resource["attributes"].get("requires", [])
                }
                for resource in resources
            },
            undefined={
                ResourceIdStr(resource["resource_id"])
                for resource in resources
                if const.ResourceState(resource["status"]) is const.ResourceState.undefined
            },
        )


class TaskManager(abc.ABC):
    """
    Interface for communication with tasks (deploy.task.Task). Offers methods to inspect intent and to report task results.
    """

    environment: uuid.UUID
    client: Client
    code_manager: CodeManager
    executor_manager: executor.ExecutorManager[executor.Executor]

    @abstractmethod
    async def get_resource_version_intent(
        self,
        resource: ResourceIdStr,
    ) -> Optional[ResourceVersionIntent]:
        """
        Returns the current version and the intent for the given resource, or None if it is not (anymore) managed by the
        scheduler.

        Acquires appropriate locks.
        """

    @abstractmethod
    async def deploy_start(
        self,
        action_id: uuid.UUID,
        resource: ResourceIdStr,
    ) -> Optional[DeployIntent]:
        """
        Register the start of deployment for the given resource and return its current version intent
        along with the last non-deploying state for its dependencies, or None if it is not (anymore)
        managed by the scheduler.

        If this method returns an intent, the caller must also call deploy_done() at the end of the deploy attempt.

        Acquires appropriate locks.
        """

    @abstractmethod
    async def deploy_done(self, deploy_intent: DeployIntent, report: executor.DeployReport) -> None:
        """
        Register the end of deployment for the given resource: update the resource state based on the deployment result
        and inform its dependencies that deployment is finished. Depending on how fresh the intent is (compared to what is
        currently managed, limited (for outdated) or no (for fully stale) scheduler state may be updated.

        Acquires appropriate locks

        :param deploy_intent: The resource's deploy intent as returned by deploy_start().
        :param report: The DeployReport object describing the result of the deployment.
        """

    @abstractmethod
    async def dryrun_done(self, report: executor.DryrunReport) -> None:
        """
        Report the result of a dry-run.
        """

    @abstractmethod
    async def fact_refresh_done(self, report: executor.GetFactReport) -> None:
        """
        Report the result of a fact refresh.
        """


class TaskRunner:
    def __init__(self, endpoint: str, scheduler: "ResourceScheduler"):
        self.endpoint = endpoint
        self.status = AgentStatus.STOPPED
        self._scheduler = scheduler
        self._task: typing.Optional[asyncio.Task[None]] = None
        self._notify_tasks: dict[uuid.UUID, asyncio.Task[None]] = {}
        # Lock to prevent race conditions on the running state of this TaskRunner
        self._notify_lock = asyncio.Lock()

    async def start(self) -> None:
        assert (
            self._task is None or self._task.done()
        ), f"Task Runner {self.endpoint} is trying to start twice, this should not happen"
        self.status = AgentStatus.STARTED
        self._task = asyncio.create_task(self._run())

    async def stop(self) -> None:
        self.status = AgentStatus.STOPPING
        for task in self._notify_tasks.values():
            if not task.done():
                task.cancel()

    async def join(self) -> None:
        assert not self.is_running(), "Joining worker that is not stopped"
        if self._task is None or self._task.done():
            return
        await self._task
        await asyncio.gather(*list(self._notify_tasks.values()), return_exceptions=True)

    async def notify(self, task_id: uuid.UUID | None = None) -> None:
        """
        Method to notify the runner that something has changed in the DB. This method will fetch the new information
        regarding the environment and the information related to the runner (agent). Depending on the desired state of the
        agent, it will either stop / start the agent or do nothing

        :param task_id: Is not None if this task was started using the `notify_sync()` method. It's used to clean up the
                        reference to the associated asyncio.Task object in `self._notify_tasks`. This value is None otherwise.
        """
        try:
            async with self._notify_lock:
                should_be_running = (
                    await self._scheduler.should_be_running()
                    and await self._scheduler.should_runner_be_running(endpoint=self.endpoint)
                )

                match self.status:
                    case AgentStatus.STARTED if not should_be_running:
                        await self.stop()
                    case AgentStatus.STOPPED if should_be_running:
                        await self.start()
                    case AgentStatus.STOPPING if should_be_running:
                        self.status = AgentStatus.STARTED
        finally:
            if task_id:
                del self._notify_tasks[task_id]

    def notify_sync(self) -> None:
        """
        Method to notify the runner that something has changed in the DB in a synchronous manner.
        """
        # We save it to be sure that the task will not be GC
        task_id = uuid.uuid4()
        task = asyncio.create_task(self.notify(task_id))
        self._notify_tasks[task_id] = task

    async def _run(self) -> None:
        """Main loop for one agent. It will first fetch or create its actual state from the DB to make sure that it's
        allowed to run."""
        while self._scheduler._running and self.status == AgentStatus.STARTED:
            work_item: work.MotivatedTask[Task] = await self._scheduler._work.agent_queues.queue_get(self.endpoint)
            try:
                await work_item.task.execute(self._scheduler, self.endpoint, work_item.reason)
            except Exception:
                LOGGER.exception(
                    "Task %s for agent %s has failed and the exception was not properly handled", work_item.task, self.endpoint
                )
            self._scheduler._work.agent_queues.task_done(self.endpoint, work_item.task)

        self.status = AgentStatus.STOPPED

    def is_running(self) -> bool:
        return self.status == AgentStatus.STARTED


class ResourceScheduler(TaskManager):
    """
    Scheduler for resource actions. Reads resource state from the database and accepts deploy, dry-run, ... requests from the
    server. Schedules these requests as tasks according to priorities and, in case of deploy tasks, requires-provides edges.

    The scheduler expects to be notified by the server whenever a new version is released.
    """

    def __init__(
        self, environment: uuid.UUID, executor_manager: executor.ExecutorManager[executor.Executor], client: Client
    ) -> None:
        """
        :param environment: the environment we work for
        :param executor_manager: the executor manager that will provide us with executors
        :param client: connection to the server
        """
        # state and work may be reassigned during initialize
        self._state: ModelState = ModelState(version=0)
        self._work: work.ScheduledWork = work.ScheduledWork(
            requires=self._state.requires.requires_view(),
            provides=self._state.requires.provides_view(),
            new_agent_notify=self._create_agent,
        )

        # We uphold two locks to prevent concurrency conflicts between external events (e.g. new version or deploy request)
        # and the task executor background tasks.
        #
        # - lock to block scheduler state access (this includes model state, scheduled work and resource timers management)
        #   during scheduler-wide state updates (e.g. trigger deploy). A single lock suffices since all state accesses
        #   (both read and write) by the task runners are short, synchronous operations (and therefore we wouldn't gain
        #   anything by allowing multiple readers).
        self._scheduler_lock: asyncio.Lock = asyncio.Lock()
        # - lock to serialize updates to the scheduler's intent (version, attributes, ...), e.g. process a new version.
        self._intent_lock: asyncio.Lock = asyncio.Lock()

        self._running = False
        # Agent name to worker task
        # here to prevent it from being GC-ed
        self._workers: dict[str, TaskRunner] = {}
        # Set of resources for which a concrete non-stale deploy is in progress, i.e. we've committed for a given intent and
        # that intent still reflects the latest resource intent
        # Apart from the obvious, this differs from the agent queues' in-progress deploys in the sense that those are simply
        # tasks that have been picked up while this set contains only those tasks for which we've already committed. For each
        # deploy task, there is a (typically) short window of time where it's considered in progress by the agent queue, but
        # it has not yet started on the actual deploy, i.e. it will still see updates to the resource intent.
        self._deploying_latest: set[ResourceIdStr] = set()
        # Set of orphaned resources for which a deploy is in progress, i.e. a deploy for a resource that went from managed to
        # unmanaged (even if it became managed again afterward, in which case it's considered a new resource that happens to
        # share the resource id)
        self._deploying_unmanaged: set[ResourceIdStr] = set()

        self.environment = environment
        self.client = client
        self.code_manager = CodeManager(client)
        self.executor_manager = executor_manager
        self.state_update_manager = ToDbUpdateManager(client, environment)

        self._timer_manager = timers.TimerManager(self)

        self._deployment_suspended: bool = False

    async def _reset(self) -> None:
        """
        Clear out all state and start empty

        only allowed when ResourceScheduler is not running
        """
        assert not self._running
        self._state.reset()
        self._work.reset()
        # Ensure we are down
        for worker in self._workers.values():
            # We have no timeout here
            # This can potentially hang when an executor hangs
            # However, the stop should forcefully kill the executor
            # This will kill the connection and free the worker
            await worker.join()
        self._workers.clear()
        self._deploying_latest.clear()
        self._deploying_unmanaged.clear()
        await self._timer_manager.reset()

    async def start(self) -> None:
        if self._running:
            return
        LOGGER.debug("Starting resource scheduler for environment %s", str(self.environment))
        await self._reset()
        await self.reset_resource_state()

        await self._initialize()

    async def stop(self) -> None:
        if not self._running:
            return
        self._running = False
        await self._timer_manager.stop()
        # Ensure workers go down
        # First stop them
        for worker in self._workers.values():
            await worker.stop()
        # Then wake them up to receive the stop
        self._work.agent_queues.send_shutdown()

    async def join(self) -> None:
        await asyncio.gather(*[worker.join() for worker in self._workers.values()])
        await self._timer_manager.join()

    async def reset_resource_state(self) -> None:
        """
        Update resources in the latest released version of the model stuck in "deploying" state.
        This can occur when the Scheduler is killed in the middle of a deployment.
        """

        await data.Resource.reset_resource_state(self.environment)

    async def load_timer_settings(self) -> None:
        """Update the timer manager after an update of the timer config"""
        await self._timer_manager.reload_config()

    async def reload_all_timers(self) -> None:
        """
        Internal, request all timers to reload. For all known resources,
        either updates or stops its associated timer, depending on its state.
        """
        # Get lock
        async with self._scheduler_lock:
            for resource, state in self._state.resource_state.items():
                deploy = Deploy(resource=resource)
                if (
                    deploy in self._work.agent_queues
                    or deploy in self._work.agent_queues.in_progress
                    or resource in self._work._waiting
                ):
                    self._timer_manager.stop_timer(resource)
                self._timer_manager.update_timer(resource, state=state)

    async def _initialize(self) -> None:
        """
        Initialize the scheduler state and continue the deployment where we were before the server was shutdown.
        Marks scheduler as running and triggers first deploys.
        """
        await self._timer_manager.initialize()
        # do not start a transaction because:
        # 1. nothing we do here before read_version is inherently transactional: the only write is atomic, and reads do not
        #   benefit from READ COMMITTED (default) isolation level.
        # 2. read_version expects to receive a connection outside of a transaction context
        async with self.state_update_manager.get_connection() as con:
            # Make sure there is an entry for this scheduler in the scheduler database table
            await data.Scheduler._execute_query(
                f"""
                    INSERT INTO {data.Scheduler.table_name()}
                    VALUES($1, NULL)
                    ON CONFLICT DO NOTHING
                """,
                self.environment,
                connection=con,
            )

            environment: Optional[data.Environment] = await data.Environment.get_by_id(self.environment, connection=con)
            assert environment is not None
            should_restore_state: bool = not typing.cast(
                bool, await environment.get(data.RESET_DEPLOY_PROGRESS_ON_START, connection=con)
            )

            # Check if we can restore the scheduler state from a previous run
            restored_state: Optional[ModelState] = (
                await ModelState.create_from_db(self.environment, connection=con) if should_restore_state else None
            )
            if restored_state is not None:
                LOGGER.debug("Scheduler initialization: restoring internal state")
                # Restore scheduler state like it was before the scheduler went down
                self._state = restored_state
                self._work = work.ScheduledWork(
                    requires=self._state.requires.requires_view(),
                    provides=self._state.requires.provides_view(),
                    new_agent_notify=self._create_agent,
                )
                restored_version: int = self._state.version
                # Set running flag because we're ready to start accepting tasks.
                # Set before scheduling first tasks because many methods (e.g. read_version) skip silently when not running
                LOGGER.debug("Scheduler initialization: resuming deploy operations and reading latest model version")
                self._running = True
                # All resources get a timer
                await self.read_version(connection=con)
                async with self._scheduler_lock:
                    LOGGER.debug("Scheduler initialization: setting up initial deploy timers")
                    self._timer_manager.update_timers(self._state.intent.keys() - self._state.dirty)

                if self._state.version == restored_version:
                    # no new version was present. Simply trigger a deploy for everything that's not in a known good state
                    LOGGER.debug("Scheduler initialization: triggering deploy")
                    await self.deploy(
                        reason="the resource scheduler was started",
                        priority=TaskPriority.INTERVAL_DEPLOY,
                    )
            else:
                # This case can occur in three different situations:
                #   1 A model version has been released, but the scheduler didn't process any version yet.
                #     In this case there is no scheduler state to restore.
                #   2 The last processed version has been deleted since the scheduler was last running
                #   3 We migrated the Inmanta server from an old version, that didn't have the resource state
                #     tracking in the database, to a version that does. To cover this case, we rely on the
                #     increment calculation to determine which resources have to be considered dirty and which
                #     not. This migration code path can be removed in a later major version.
                #
                # In cases 1 and 2, all resources are expected to be in the increment (scheduler hasn't processed any versions
                # means we haven't ever deployed anything yet), so those could be covered by simply reading in only the latest
                # version, or by keeping state as is and calling into normal read_version() flow. However, while the migration
                # path is still required for backwards compatibility (3) anyway, we unifi the three cases for simplicity.

                # Set running flag because we're ready to start accepting tasks.
                # Set before scheduling first tasks because many methods (e.g. read_version) skip silently when not running
                LOGGER.debug("Scheduler initialization: resuming deploy operations and reading latest model version")
                self._running = True
                await self._recover_scheduler_state_using_increments_calculation(connection=con)
            LOGGER.debug("Scheduler initialization: finished initialization, scheduler is fully up and running")

    async def _recover_scheduler_state_using_increments_calculation(self, *, connection: asyncpg.connection.Connection) -> None:
        """
        This method exists for backwards compatibility reasons. It initializes the scheduler state
        by relying on the increments calculation logic. This method starts the deployment process.

        :param connection: Connection to use for db operations. Should not be in a transaction context.
        """
        try:
            model: ModelVersion = await self._get_single_model_version_from_db(connection=connection)
        except KeyError:
            # No model version has been released yet.
            return
        LOGGER.debug("Read latest model version %d", model.version)
        # Rely on the incremental calculation to determine which resources should be deployed and which not.
        up_to_date_resources: Set[ResourceIdStr]
        up_to_date_resources, last_deploy_time = await ConfigurationModel.get_last_deployed_and_neg_increment(
            self.environment, model.version, connection=connection
        )
        await self._new_version(
            [model],
            up_to_date_resources=up_to_date_resources,
            last_deploy_time=last_deploy_time,
            reason="the scheduler was started",
            connection=connection,
        )

    async def deploy(
        self,
        *,
        reason: str,
        priority: TaskPriority = TaskPriority.USER_DEPLOY,
        agent: Optional[str] = None,
    ) -> None:
        """
        Trigger a deploy

        :param agent: If given, deploy resources only for this agent. Otherwise deploy for all agents.
        """
        if not self._running:
            LOGGER.debug("Ignoring deploy request for halted resource scheduler")
            return
        async with self._scheduler_lock:
            to_deploy: Set[ResourceIdStr] = (
                self._state.dirty if agent is None else self._state.dirty & self._state.resources_by_agent.get(agent, set())
            )
            if agent is not None:
                LOGGER.debug("Triggering deploy for %d resources on agent %s because %s", len(to_deploy), agent, reason)
            else:
                LOGGER.debug("Triggering deploy for %d resources because %s", len(to_deploy), reason)
            self._timer_manager.stop_timers(to_deploy)
            self._work.deploy_with_context(to_deploy, reason=reason, priority=priority, deploying=self._deploying_latest)

    async def repair(
        self,
        *,
        reason: str,
        priority: TaskPriority = TaskPriority.USER_REPAIR,
        agent: Optional[str] = None,
    ) -> None:
        """
        Trigger a repair, i.e. mark all unblocked resources as dirty, then trigger a deploy.

        :param agent: If given, repair resources only for this agent. Otherwise repair for all agents.
        """

        def should_deploy_resource(resource: ResourceIdStr) -> bool:
            if (resource_state := self._state.resource_state.get(resource)) is not None:
                # For now, we repair even resources marked as TEMPORARILY_BLOCKED, just in case our assumptions are wrong
                # We will relax this once we have more confidence in the correct tracking of the state (#8580)
                return resource_state.blocked is not Blocked.BLOCKED
            # No state was found for this resource. Should probably not happen
            # but err on the side of caution and mark for redeploy.
            return True

        if not self._running:
            LOGGER.debug("Ignoring repair request for halted resource scheduler")
            return
        async with self._scheduler_lock:
            in_scope: Set[ResourceIdStr] = (
                self._state.intent.keys() if agent is None else self._state.resources_by_agent.get(agent, set())
            )

            to_deploy: Set[ResourceIdStr] = {resource for resource in in_scope if should_deploy_resource(resource)}
            if agent is not None:
                LOGGER.debug("Triggering deploy for %d resources on agent %s because %s", len(to_deploy), agent, reason)
            else:
                LOGGER.debug("Triggering deploy for %d resources because %s", len(to_deploy), reason)
            self._state.dirty.update(to_deploy)
            self._timer_manager.stop_timers(to_deploy)
            self._work.deploy_with_context(to_deploy, reason=reason, priority=priority, deploying=self._deploying_latest)

    async def dryrun(self, dry_run_id: uuid.UUID, version: int) -> None:
        if not self._running:
            LOGGER.debug("Ignoring dry-run request for halted resource scheduler")
            return

        paused_agents = await self.all_paused_agents()

        LOGGER.debug("Triggering dry-run %s for version %d", str(dry_run_id), version)
        model: ModelVersion = await self._get_single_model_version_from_db(version=version)
        for resource, resource_intent in model.resources.items():
            if resource in model.undefined:
                continue

            if resource_intent.id.agent_name in paused_agents:
                # Paused agents are handled on the calling side
                continue

            self._work.agent_queues.queue_put_nowait(
                DryRun(
                    resource=resource_intent.resource_id,
                    version=model.version,
                    resource_intent=ResourceIntent(
                        resource_id=resource_intent.resource_id,
                        attribute_hash=resource_intent.attribute_hash,
                        attributes=resource_intent.attributes,
                    ),
                    dry_run_id=dry_run_id,
                ),
                priority=TaskPriority.DRYRUN,
            )

    async def get_facts(self, resource: dict[str, object]) -> None:
        if not self._running:
            return
        rid = Id.parse_id(resource["id"]).resource_str()
        self._work.agent_queues.queue_put_nowait(
            RefreshFact(resource=rid),
            priority=TaskPriority.FACT_REFRESH,
        )

    async def deploy_resource(self, resource: ResourceIdStr, reason: str, priority: TaskPriority) -> None:
        """
        Make sure the given resource is marked for deployment with at least the provided priority.
        If the given priority is higher than the previous one (or if it didn't exist before), a new deploy
        task will be scheduled with the provided reason.
        """
        async with self._scheduler_lock:
            if resource not in self._state.resource_state:
                # The resource was removed from the model by the time this method was triggered
                return

            # When explicitly requested for a single resource, we allow deploying even deploys marked as
            # Blocked.TEMPORARILY_BLOCKED
            if self._state.resource_state[resource].blocked is Blocked.BLOCKED:  # Can't deploy
                return
            self._timer_manager.stop_timer(resource)
            self._work.deploy_with_context(
                {resource},
                reason=reason,
                priority=priority,
                deploying=self._deploying_latest,
            )

    async def _get_single_model_version_from_db(
        self,
        *,
        version: int | None = None,
        connection: Optional[asyncpg.connection.Connection] = None,
    ) -> ModelVersion:
        """
        Returns a single model version as fetched from the database. Returns the requested version if provided
        (regardless of whether it has been released or not), otherwise returns the latest released version.

        Raises KeyError if no specific version has been provided and no released versions exist.
        """
        async with self.state_update_manager.get_connection(connection) as con:
            if version is None:
                # Fetch the latest released model version
                cm_version = await ConfigurationModel.get_latest_version(self.environment, connection=con)
                if cm_version is None:
                    raise KeyError()
                version = cm_version.version

            resources_from_db = await data.Resource.get_resources_for_version_raw(
                environment=self.environment,
                version=version,
                projection=ResourceRecord.__required_keys__,
                connection=con,
            )

            return ModelVersion.from_db_records(
                version,
                resources_from_db,  # type: ignore
            )

    async def read_version(
        self,
        *,
        connection: Optional[asyncpg.connection.Connection] = None,
    ) -> None:
        """
        Update model state and scheduled work based on the latest released version in the database,
        e.g. when a new version is released. Triggers a deploy after updating internal state:
        - schedules new or updated resources to be deployed
        - schedules any resources that are not in a known good state.
        - rearranges deploy tasks by requires if required

        :param connection: Connection to use for db operations. Should not be in a transaction context.
        """
        if not self._running:
            LOGGER.debug("Ignoring request to read latest version for halted resource scheduler")
            return
        async with self._intent_lock, self.state_update_manager.get_connection(connection) as con:
            # Note: we're not very sensitive to races on the latest released version here. The server will always notify us
            # *after* a new version is released. So we'll always be in one of two scenearios:
            # - the latest released version was released before we started processing this notification
            # - a new version was released after we started processing this communication or is being released now, in which
            #   case a new notification will be / have been sent and blocked on the intent locked until we're done here.
            # So if we end up with a race, we can be confident that we'll always process the associated notification soon.
            resources_by_version: Sequence[tuple[int, Sequence[Mapping[str, object]]]] = (
                await data.Resource.get_resources_since_version_raw(
                    self.environment,
                    since=self._state.version,
                    projection=ResourceRecord.__required_keys__,
                    connection=con,
                )
            )
            new_versions: Sequence[ModelVersion] = [
                ModelVersion.from_db_records(
                    version,
                    resources,  # type: ignore
                )
                for version, resources in resources_by_version
            ]
            LOGGER.debug(
                "Read %d model versions newer than the currently managed version %d",
                len(resources_by_version),
                self._state.version,
            )

            await self._new_version(
                new_versions,
                reason="a new version was released",
                connection=con,
            )

    async def _get_intent_changes(
        self,
        new_versions: Sequence[ModelVersion],
        *,
        up_to_date_resources: Optional[Set[ResourceIdStr]] = None,
    ) -> tuple[ModelVersion, dict[ResourceIdStr, ResourceIntentChange]]:
        """
        Returns a consolidated overview of changes to resource intent, given a series of model versions to process.

        Returns a tuple of
        - the resulting model version after processing all versions.
        - a mapping of resources to their change of intent, after processing all versions in sequence. Only contains resources
            with a change of intent. May include resources with DELETED change of intent even if they aren't part of the current
            state.

        :param new_versions: The new versions to process, in ascending order.
        :param up_to_date_resources: A set of resources that are considered up to date and in a known good state, regardless
            of the current state information.
        """
        up_to_date_resources = set() if up_to_date_resources is None else up_to_date_resources

        if not new_versions:
            raise ValueError("Expected at least one new model version")

        LOGGER.debug("Consolidating changes from %d new versions", len(new_versions))

        version: int
        intent: dict[ResourceIdStr, ResourceIntent] = {}
        resource_requires: dict[ResourceIdStr, Set[ResourceIdStr]] = {}
        # keep track of all new resource intent and all changes of intent relative to the currently managed version
        intent_changes: dict[ResourceIdStr, ResourceIntentChange] = {}
        # all undefined resources in the new model versions. Newer version information overrides older ones.
        undefined: set[ResourceIdStr] = set()

        previous_version: int = self._state.version
        for model in new_versions:
            version = model.version
            # resources that don't exist anymore in this version
            for resource in (self._state.intent.keys() | intent.keys()) - model.resources.keys():
                with contextlib.suppress(KeyError):
                    del intent[resource]
                    del resource_requires[resource]
                if resource not in intent_changes or intent_changes[resource].change is not ResourceIntentChangeType.DELETED:
                    intent_changes[resource] = Deleted(last_seen_version=previous_version)
                undefined.discard(resource)

            i: int = 0
            for resource, resource_intent in model.resources.items():
                # this loop is race-free, potentially slow, and completely synchronous
                # => regularly pass control to the event loop to not block scheduler operation during update prep
                if i >= NB_ITERATIONS_PASS_IO_LOOP:
                    await asyncio.sleep(0)
                    i = 0
                i += 1

                # register new resource intent and requires, overriding previous information
                intent[resource] = resource_intent
                # we only care about the requires-provides changes relative to the currently managed ones,
                # not relative to any of the older versions => simply override
                resource_requires[resource] = model.requires.get(resource, set())

                # determine change of intent
                # no change of intent, by definition
                if resource in up_to_date_resources:
                    continue
                # new resources
                if (
                    # exists in this version but not in managed version
                    resource not in self._state.intent.keys()
                    # deleted in a previously processed version and reappeared now
                    # => consider as a new resource rather than an update
                    or (
                        resource in intent_changes
                        and intent_changes[resource].change in (ResourceIntentChangeType.DELETED, ResourceIntentChangeType.NEW)
                    )
                ):
                    intent_changes[resource] = New()
                # resources we already manage
                elif resource_intent.attribute_hash != self._state.intent[resource].attribute_hash:
                    intent_changes[resource] = Updated()
                # no change of intent for this resource, unless defined status changed

                # determine new defined status
                is_undefined: bool
                if resource in model.undefined:
                    undefined.add(resource)
                    is_undefined = True
                else:
                    undefined.discard(resource)
                    is_undefined = False
                if resource not in intent_changes and is_undefined != (
                    self._state.resource_state[resource].compliance is Compliance.UNDEFINED
                ):
                    # resource's defined status changed
                    intent_changes[resource] = Updated()

            previous_version = version

        return (
            ModelVersion(
                version=version,
                resources=intent,
                requires=resource_requires,
                undefined=undefined,
            ),
            intent_changes,
        )

    async def _new_version(
        self,
        new_versions: Sequence[ModelVersion],
        *,
        up_to_date_resources: Optional[Set[ResourceIdStr]] = None,
        last_deploy_time: Mapping[ResourceIdStr, datetime.datetime] | None = None,
        reason: str = "a new version was released",
        connection: Optional[asyncpg.connection.Connection] = None,
    ) -> None:
        """
        Register one or more newly released versions with the scheduler. Updates scheduled state and scheduled work accordingly.

        Expects to be called under the intent lock, and with monotonically increasing versions (relative to the previous calls
        as well as within the new_versions sequence).

        :param up_to_date_resources: Set of resources that are to be considered in an assumed good state due to a previously
            successful deploy for the same intent. Should not include any blocked resources. Mostly intended for the very first
            version when the scheduler is started with a fresh state.
        :param last_deploy_time: last known deploy time of all resources. only intended for the very first
            version when the scheduler is started with a fresh state.
            It must contain ALL resources as it is used to start the timers
        :param reason: The reason to associate with any deploys caused by this newly released version.
        :param connection: Connection to use for db operations. Should not be in a transaction context for two reasons:
            1. This method acquires the scheduler lock, and needs to have control about how it is allowed to interleave with
                transaction commit, if at all.
            2. This method may have a long pre-processing phase, it opens a transaction itself when appropriate
        """
        if not new_versions:
            return
        if connection is not None and connection.is_in_transaction():
            raise ValueError("_new_version() expects its connection to not be in a transaction context")

        first_version: bool = self._state.version == 0

        up_to_date_resources = set() if up_to_date_resources is None else up_to_date_resources
        last_deploy_time = {} if last_deploy_time is None else last_deploy_time

        # pre-process new model versions before acquiring the scheduler lock.
        model: ModelVersion
        intent_changes: Mapping[ResourceIdStr, ResourceIntentChange]
        model, intent_changes = await self._get_intent_changes(new_versions, up_to_date_resources=up_to_date_resources)

        LOGGER.debug(
            (
                "Processing intent changes for %d resources between currently managed model version %d"
                " and new model version %d, because %s"
            ),
            len(intent_changes),
            self._state.version,
            model.version,
            reason,
        )

        # Track potential changes in requires per resource
        added_requires: dict[ResourceIdStr, Set[ResourceIdStr]] = {}  # includes new resources if they have at least one req
        dropped_requires: dict[ResourceIdStr, Set[ResourceIdStr]] = {}
        i: int = 0
        for resource, requires in model.requires.items():
            old_requires: Set[ResourceIdStr] = self._state.requires.get(resource, set())
            added = requires - old_requires
            if added:
                added_requires[resource] = added
            removed = old_requires - requires
            if removed:
                dropped_requires[resource] = removed

            # pass control back to event loop to maintain scheduler operations until we acquire the lock
            if i >= NB_ITERATIONS_PASS_IO_LOOP:
                await asyncio.sleep(0)
                i = 0
            i += 1

        # convert intent changes to sets for bulk processing
        deleted: dict[ResourceIdStr, Deleted] = {}
        new: set[ResourceIdStr] = set()
        updated: set[ResourceIdStr] = set()
        # resources that were previously undefined but not anymore, including those considered new if they match this property
        became_defined: set[ResourceIdStr] = set()
        # resources that are now undefined but weren't before, including new resources if they are undefined
        became_undefined: set[ResourceIdStr] = set()
        i = 0
        for resource, change in intent_changes.items():
            # pass control back to event loop to maintain scheduler operations until we acquire the lock
            if i >= NB_ITERATIONS_PASS_IO_LOOP:
                await asyncio.sleep(0)
                i = 0
            i += 1

            match change:
                case Deleted():
                    deleted[resource] = change
                    continue
                case New():
                    new.add(resource)
                case Updated():
                    updated.add(resource)
                case _ as _never:
                    typing.assert_never(_never)
            resource_state: Optional[ResourceState] = self._state.resource_state.get(resource)
            if resource_state is None:
                if resource in model.undefined:
                    became_undefined.add(resource)
                continue
            attribute_hash_changed: bool = (
                model.resources[resource].attribute_hash != self._state.intent[resource].attribute_hash
            )
            attribute_hash_unchanged_warning_fmt: str = (
                "The resource with id %s has become %s, but the hash has not changed."
                " This may lead to unexpected deploy behavior"
            )
            if resource in model.undefined and resource_state.compliance is not Compliance.UNDEFINED:
                if not attribute_hash_changed:
                    LOGGER.warning(attribute_hash_unchanged_warning_fmt, resource, "undefined")
                became_undefined.add(resource)
            elif resource not in model.undefined and resource_state.compliance is Compliance.UNDEFINED:
                if not attribute_hash_changed:
                    LOGGER.warning(attribute_hash_unchanged_warning_fmt, resource, "defined")
                became_defined.add(resource)

        # resources that are to be considered new, even if they are already being managed
        force_new: Set[ResourceIdStr] = self._state.intent.keys() & new

        # assert invariants of the constructed sets
        assert len(intent_changes) == len(deleted.keys() | new | updated) == len(deleted) + len(new) + len(updated)
        assert len(became_defined | became_undefined) == (len(became_defined) + len(became_undefined))

        # pass control to IO loop once more before we acquire the lock
        await asyncio.sleep(0)

        # in the current implementation everything below the lock is synchronous, it is however still required (1)
        # and desired even if it weren't strictly required (2):
        # 1. other operations under lock may pass control back to the event loop under the lock. Meaning that even if this
        #   method itself were fully synchronous, it will always still be called in some async context, which may interleave
        #   with awaits in other operations under lock.
        #   and has to be even apart from motivations 2-3, it may interleave with
        # 2. clarity: it clearly signifies that this is the atomic and performance-sensitive part
        async with self._scheduler_lock:
            # update model version
            self._state.version = model.version
            # update resource intent
            for resource in up_to_date_resources:
                # Registers resource and removes from the dirty set
                self._state.update_resource(
                    model.resources[resource], known_compliant=True, last_deployed=last_deploy_time.get(resource, None)
                )
            for resource in new | updated:
                # update resource state and dirty set
                self._state.update_resource(
                    model.resources[resource],
                    force_new=intent_changes[resource].change is ResourceIntentChangeType.NEW,
                    undefined=resource in model.undefined,
                    last_deployed=last_deploy_time.get(resource, None),
                )
            # update requires
            for resource in added_requires.keys() | dropped_requires.keys():
                self._state.update_requires(resource, model.requires[resource])

            # update transitive state
            transitive_unblocked, transitive_blocked = self._state.update_transitive_state(
                new_undefined=became_undefined,
                verify_blocked=added_requires.keys(),
                verify_unblocked=became_defined | dropped_requires.keys(),
            )
            # update TEMPORARILY_BLOCKED (skipped-for-dependencies) state
            # for resources with a dependency for which state was reset
            resources_with_reset_requires: Set[ResourceIdStr] = set(
                itertools.chain.from_iterable(
                    self._state.requires.provides_view().get(resource, set()) for resource in force_new
                )
            )
            for resource in resources_with_reset_requires:
                if (
                    # it is currently TEMPORARILY_BLOCKED blocked
                    self._state.resource_state[resource].blocked is Blocked.TEMPORARILY_BLOCKED
                    # it shouldn't be any longer
                    and not self._state.should_skip_for_dependencies(resource)
                ):
                    self._state.resource_state[resource].blocked = Blocked.NOT_BLOCKED

            # Update set of in-progress deploys that became unmanaged
            self._deploying_unmanaged.update(self._deploying_latest & (new | deleted.keys()))
            # Update set of in-progress non-stale deploys by trimming resources with new state
            self._deploying_latest.difference_update(intent_changes.keys(), transitive_blocked)

            # Remove timers for resources that are:
            #    - in the dirty set (because they will be picked up by the scheduler eventually)
            #    - blocked: must not be deployed
            #    - deleted from the model
            self._timer_manager.stop_timers(self._state.dirty | became_undefined | transitive_blocked)
            self._timer_manager.remove_timers(deleted.keys())
            # Install timers for initial up-to-date resources. They are up-to-date now,
            # but we want to make sure we periodically repair them.
            self._timer_manager.update_timers(
                up_to_date_resources | ((transitive_unblocked | resources_with_reset_requires) - self._state.dirty)
            )

            # ensure deploy for ALL dirty resources, not just the new ones
            self._work.deploy_with_context(
                self._state.dirty,
                reason=reason,
                priority=TaskPriority.NEW_VERSION_DEPLOY,
                deploying=self._deploying_latest,
                added_requires=added_requires,
                dropped_requires=dropped_requires,
            )
            for resource in deleted:
                if resource in self._state.resource_state:
                    self._state.drop(resource)  # Removes from the dirty set
            for resource in deleted.keys() | became_undefined | transitive_blocked:
                self._work.delete_resource(resource)

            # Updating the blocked state should be done under the scheduler lock, because this state is written
            # by both the deploy and the new version code path.
            resources_with_updated_blocked_state: Set[ResourceIdStr] = (
                became_undefined | transitive_blocked | transitive_unblocked
            )

        LOGGER.debug(
            "Successfully updated resource scheduler to manage model version %d. Writing intent changes to the database.",
            model.version,
        )
        async with self.state_update_manager.get_connection(connection=connection) as con, con.transaction():
            # Update intent for resources with new desired state
            # Safe to update outside of the lock: scheduler persisted intent is allowed to lag behind its in-memory intent
            # because we always update again after recovering, and we never write to these same columns from any other place.
            # TODO[#8541]: with the changes planned in #8541, the "never write to these same columns from any other places"
            #   statement may not hold anymore. In any case, the transaction must not interleave or wrap the scheduler lock.
            await self.state_update_manager.update_resource_intent(
                self.environment,
                intent={
                    rid: (self._state.resource_state[rid], self._state.intent[rid])
                    for rid in (new | updated | resources_with_updated_blocked_state) - deleted.keys()
                },
                update_blocked_state=True,
                connection=con,
            )
            # Mark orphaned resources
            if first_version:
                # We're starting fresh. Make sure to mark all orphans, because we may be skipping some unprocessed versions.
                await self.state_update_manager.mark_all_orphans(
                    self.environment, current_version=model.version, connection=con
                )
            else:
                # We're processing versions relative to an already processed version => deleted contains all orphans.
                await self.state_update_manager.mark_as_orphan(self.environment, deleted.keys(), connection=con)
            await self.state_update_manager.set_last_processed_model_version(
                self.environment, self._state.version, connection=con
            )
        LOGGER.debug("Finished writing changes for model version %d to the database", model.version)

    def _create_agent(self, agent: str) -> None:
        """Start processing for the given agent"""
        self._workers[agent] = TaskRunner(endpoint=agent, scheduler=self)
        self._workers[agent].notify_sync()

    async def should_be_running(self) -> bool:
        """
        Check in the DB (authoritative entity) if the Scheduler should be running
            i.e. if the environment is not halted.
        """
        current_environment = await Environment.get_by_id(self.environment)
        assert current_environment
        return not current_environment.halted

    async def should_runner_be_running(self, endpoint: str) -> bool:
        """
        Check in the DB (authoritative entity) if the agent (or the Scheduler if endpoint == Scheduler id) should be running
            i.e. if it is not paused.

        :param endpoint: The name of the agent
        """
        await data.Agent.insert_if_not_exist(environment=self.environment, endpoint=endpoint)
        if self._deployment_suspended:
            return False
        current_agent = await data.Agent.get(env=self.environment, endpoint=endpoint)
        return not current_agent.paused

    async def all_paused_agents(self) -> set[str]:
        return {agent.name for agent in await data.Agent.get_list(environment=self.environment, paused=True)}

    async def refresh_agent_state_from_db(self, name: str) -> None:
        """
        Refresh from the DB (authoritative entity) the actual state of the agent.
            - If the agent is not paused: It will make sure that the agent is running.
            - If the agent is paused: Stop a particular agent.

        :param name: The name of the agent
        """
        if name in self._workers:
            await self._workers[name].notify()

    async def refresh_all_agent_states_from_db(self) -> None:
        """
        Refresh from the DB (authoritative entity) the actual state of all agents.
            - If an agent is not paused: It will make sure that the agent is running.
            - If an agent is paused: Stop a particular agent.
        """
        for worker in self._workers.values():
            await worker.notify()

    async def is_agent_running(self, name: str) -> bool:
        """
        Return True if the provided agent is running, at least an agent that the Scheduler is aware of

        :param name: The name of the agent
        """
        return name in self._workers and self._workers[name].is_running()

    # TaskManager interface

    def _get_resource_intent(self, resource: ResourceIdStr) -> Optional[ResourceIntent]:
        """
        Get intent of a given resource.
        Always expected to be called under lock
        """
        try:
            return self._state.intent[resource]
        except KeyError:
            # Stale resource
            # May occur in rare races between new_version and acquiring the lock we're under here. This race is safe
            # because of this check, and an intrinsic part of the locking design because it's preferred over wider
            # locking for performance reasons.
            return None

    async def get_resource_version_intent(self, resource: ResourceIdStr) -> Optional[ResourceVersionIntent]:
        async with self._scheduler_lock:
            # fetch resource intent under lock
            resource_intent = self._get_resource_intent(resource)
            if resource_intent is None:
                return None
            return ResourceVersionIntent(
                model_version=self._state.version,
                intent=resource_intent,
                all_types_for_agent=list(self._state.types_per_agent[self._state.intent[resource].id.agent_name]),
            )

    async def deploy_start(self, action_id: uuid.UUID, resource: ResourceIdStr) -> Optional[DeployIntent]:
        async with self._scheduler_lock:
            # fetch resource intent under lock
            resource_intent = self._get_resource_intent(resource)
            if resource_intent is None or self._state.resource_state[resource].blocked is Blocked.BLOCKED:
                # We are trying to deploy a stale resource.
                return None
            dependencies = await self._get_last_non_deploying_state_for_dependencies(resource=resource)
            self._deploying_latest.add(resource)
            deploy_intent = DeployIntent(
                model_version=self._state.version,
                intent=resource_intent,
                dependencies=dependencies,
                deploy_start=datetime.datetime.now().astimezone(),
                all_types_for_agent=list(self._state.types_per_agent[self._state.intent[resource].id.agent_name]),
            )
            # Update the state in the database.
            await self.state_update_manager.send_in_progress(
                action_id, Id.parse_id(ResourceVersionIdStr(f"{resource},v={self._state.version}"))
            )
            return deploy_intent

    async def deploy_done(self, deploy_intent: DeployIntent, report: executor.DeployReport) -> None:
        finished = datetime.datetime.now().astimezone()
        try:
            state: Optional[ResourceState]
            try:
                state = await self._update_scheduler_state_for_finished_deploy(deploy_intent, report, finished)
            except StaleResource:
                # The resource is no longer managed (or in rare cases it has shortly become unmanaged sometime
                # between this version and the currently managed version, either way, the deploy that finished
                # represents a stale intent). We do still want to report that the deploy finished for this version,
                # even if it is stale, we simply don't report any state. new, ununtouched state for this resource
                # id), because that still represents the most current state for this resource, even if it is
                # independent from this deploy. The scheduler will then report that this resource id with this
                # specific version finished deploy.
                state = None
            # Write deployment result to the database.
            await self.state_update_manager.send_deploy_done(
                attribute_hash=deploy_intent.intent.attribute_hash,
                result=report,
                state=state,
                started=deploy_intent.deploy_start,
                finished=finished,
            )
        finally:
            # Always do this, even if the DB is broken
            async with self._scheduler_lock:
                # report to the scheduled work that we're done
                self._work.finished_deploy(report.resource_id)
                state = self._state.resource_state.get(deploy_intent.intent.resource_id)
                if state is not None:
                    self._timer_manager.update_timer(deploy_intent.intent.resource_id, state=state)

    async def dryrun_done(self, report: executor.DryrunReport) -> None:
        await self.state_update_manager.dryrun_update(env=self.environment, dryrun_result=report)

    async def fact_refresh_done(self, report: executor.GetFactReport) -> None:
        await self.state_update_manager.set_parameters(fact_result=report)

    async def _update_scheduler_state_for_finished_deploy(
        self, deploy_intent: DeployIntent, result: executor.DeployReport, finished: datetime.datetime
    ) -> ResourceState:
        """
        Update the state of the scheduler based on the DeployResult of the given resource.

        May add log messages to the given deploy result object.

        :raise StaleResource: This update is about a resource that is no longer managed by the server.
        :return: The new state of the resource, even if no changes were made. The returned object is a static copy that
            represents the state at the end of the deploy, and can therefore safely be returned out of the scheduler lock.
        """
        resource: ResourceIdStr = result.resource_id
        deploy_result: DeployResult = DeployResult.from_handler_resource_state(result.resource_state)

        async with self._scheduler_lock:
            # refresh resource intent for latest model state
            resource_intent: Optional[ResourceIntent] = self._state.intent.get(resource, None)

            if resource_intent is None or resource in self._deploying_unmanaged:
                # we are stale and removed
                self._deploying_unmanaged.discard(resource)
                raise StaleResource()

            state: ResourceState = self._state.resource_state[resource]

            recovered_from_failure: bool = deploy_result is DeployResult.DEPLOYED and state.last_deploy_result not in (
                DeployResult.DEPLOYED,
                DeployResult.NEW,
            )

            # The second part of the or would not be required because is implied by the first,
            # except that we don't enforce the hash diff.
            # We emit a warning if we observe this, but that still doesn't prevent it.
            # While it should not happen it can
            if (
                resource_intent.attribute_hash != deploy_intent.intent.attribute_hash
                or state.compliance is Compliance.UNDEFINED
            ):
                # We are stale but still the last deploy
                # We can update the last_deploy_result (which is about last deploy)
                # We can't update compliance (which is about active state only)
                # None of the event propagation or other update happen either for the same reason
                # except for the event to notify dependents of failure recovery (to unblock skipped for dependencies)
                # because we might otherwise miss the recovery (in the sense that the next deploy wouldn't be a transition
                # from a bad to a good state, since we're transitioning to that good state now).

                state.last_deploy_result = deploy_result
                state.last_deployed = finished
                if recovered_from_failure:
                    self._send_events(resource_intent, stale_deploy=True, recovered_from_failure=True)
                return state.copy()

            # We are not stale
            state.compliance = Compliance.COMPLIANT if deploy_result is DeployResult.DEPLOYED else Compliance.NON_COMPLIANT

            # first update state, then send out events
            self._deploying_latest.remove(resource)
            state.last_deploy_result = deploy_result
            state.last_deployed = finished

            # Check if we need to mark a resource as temporarily blocked
            # We only do that if it is not already blocked (Blocked.BLOCKED)
            # We might already be unblocked if a dependency succeeded on another agent, e.g. while waiting for the lock
            # so HandlerResourceState.skipped_for_dependency might be outdated, we have an inconsistency between the
            # state of the dependencies and the exception that was raised by the handler.
            # If all dependencies are compliant we don't want to temporarily block this resource.
            if (
                state.blocked is not Blocked.BLOCKED
                and result.resource_state is const.HandlerResourceState.skipped_for_dependency
                and self._state.should_skip_for_dependencies(resource)
            ):
                state.blocked = Blocked.TEMPORARILY_BLOCKED
                # Remove this resource from the dirty set when we block it
                self._state.dirty.discard(resource)
            elif deploy_result is DeployResult.DEPLOYED:
                # Remove this resource from the dirty set if it is successfully deployed
                self._state.dirty.discard(resource)
                if state.blocked is Blocked.TEMPORARILY_BLOCKED:
                    # For now, we make sure to schedule even TEMPORARILY_BLOCKED resources for repair, just in case we have made
                    # incorrect assumptions. If this happens, we mark it as unblocked and we trigger a warning.
                    state.blocked = Blocked.NOT_BLOCKED

                    log_line: data.LogLine = self._get_temporarily_blocked_warning(deploy_intent)
                    # write to resource action (and scheduler) log
                    log_line.write_to_logger_for_resource(
                        agent=deploy_intent.intent.id.agent_name,
                        resource_version_string=deploy_intent.intent.id.copy(
                            version=deploy_intent.model_version
                        ).resource_version_str(),
                    )
                    # write to database (via send_deploy_done())
                    result.messages.append(log_line)
            else:
                # In most cases it will already be marked as dirty but in rare cases the deploy that just finished might
                # have been triggered by an event, on a previously successful deployed resource. Either way, a failure
                # (or skip) causes it to become dirty now.
                self._state.dirty.add(resource)

            # propagate events
            self._send_events(resource_intent, recovered_from_failure=recovered_from_failure)

            return state.copy()

    def _send_events(
        self,
        sending_resource: ResourceIntent,
        *,
        stale_deploy: bool = False,
        recovered_from_failure: bool,
    ) -> None:
        """
        Send events to the appropriate dependents of the given resources. Sends out normal events to declared event listeners
        unless this was triggered by a stale deploy. Additionally, if this was triggered by a failure recovery, it unblocks
        skipped dependents where appropriate and sends them an event to deploy.

        Expects to be called under the scheduler lock.

        :param sending_resource: Details for the given resource that has just finished deploying.
        :param state_deploy: The events are triggered by a stale deploy. Only recovery events will be sent.
        :param recovered_from_failure: This resource went from 'not deployed' to 'deployed',
            inform unblocked dependants that they might be able to progress.
        """
        resource_id: ResourceIdStr = sending_resource.resource_id
        provides: Set[ResourceIdStr] = self._state.requires.provides_view().get(resource_id, set())

        event_listeners: Set[ResourceIdStr]
        recovery_listeners: Set[ResourceIdStr]

        if stale_deploy or not sending_resource.attributes.get(const.RESOURCE_ATTRIBUTE_SEND_EVENTS, False):
            event_listeners = set()
        else:
            event_listeners = {
                dependent
                for dependent in provides
                if (dependent_intent := self._state.intent.get(dependent, None)) is not None
                if self._state.resource_state[dependent].blocked is Blocked.NOT_BLOCKED
                # default to True for backward compatibility, i.e. not all resources have the field
                if dependent_intent.attributes.get(const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS, True)
            }

        if not recovered_from_failure:
            recovery_listeners = set()
        else:
            recovery_listeners = {
                dependent
                for dependent in provides
                if self._state.resource_state[dependent].blocked is Blocked.TEMPORARILY_BLOCKED
            }
            # These resources might be able to progress now -> unblock them in addition to sending the event
            self._state.dirty.update(recovery_listeners)
            for skipped_dependent in recovery_listeners:
                # TODO[#8541]: persist in database
                self._state.resource_state[skipped_dependent].blocked = Blocked.NOT_BLOCKED

        all_listeners: Set[ResourceIdStr] = event_listeners | recovery_listeners
        if all_listeners:
            # do not pass deploying tasks because for event propagation we really want to start a new one,
            # even if the current intent is already being deployed
            task = Deploy(resource=resource_id)
            assert task in self._work.agent_queues.in_progress
            priority = self._work.agent_queues.in_progress[task]
            self._timer_manager.stop_timers(all_listeners)
            self._work.deploy_with_context(
                all_listeners,
                reason=(
                    f"a recovery event was received from {resource_id}"
                    if recovered_from_failure
                    else f"an event was received from {resource_id}"
                ),
                priority=priority,
                deploying=self._deploying_latest,
                # force a new deploy to be scheduled because ongoing deploys can not capture the event.
                # We desire new deploys to be scheduled, even if any are ongoing for the same intent.
                force_deploy=True,
            )

    def _get_temporarily_blocked_warning(self, deploy_intent: DeployIntent) -> data.LogLine:
        """
        Warn the user about a resource that was marked as TEMPORARILY_BLOCKED that deployed successfully. Does not actually log
        anything, only returns a log line object for the resource action log.

        Should only be called when the resource did in fact deploy when it was expected to skip for dependencies again.
        Inspects dependencies' state to determine the most appropriate warning message (handler bug or scheduler bug).

        :param deploy_intent: The intent that was just deployed successfully.
        """
        bad_dependencies: Mapping[ResourceIdStr, const.ResourceState] = {
            dependency: dependency_state
            for dependency, dependency_state in deploy_intent.dependencies.items()
            if dependency_state != const.ResourceState.deployed
        }
        if bad_dependencies:
            return data.LogLine.log(
                logging.WARNING,
                (
                    "Resource %(resource)s was expected to skip for dependencies but it deployed successfully."
                    " The handler for this resource raised a SkipResourceForDependencies exception in a previous"
                    " execution, indicating that it would only be able to progress once all requires reached a"
                    " deployed state. Some requires (%(dependencies)s) are still in a non-deployed state,"
                    " therefore it was assumed that a new deploy attempt would have no effect."
                    " This indicates incorrect usage of the exception, and it will lead to resources getting stuck"
                    " in the skipped state. While this can be worked around by triggering a repair for now, that"
                    " may not work in the future."
                    " Please check your handler implementation, and make sure to raise the generic SkipResource"
                    " rather than SkipResourceForDependencies if you wish to skip a deploy for any other reason"
                    " than to wait until all requires are in a good state."
                    " Please contact support if you believe your handler implementation is correct after all."
                ),
                resource=deploy_intent.intent.resource_id,
                dependencies=", ".join(f"{r}: {state.name}" for r, state in bad_dependencies.items()),
            )
        else:
            return data.LogLine.log(
                logging.WARNING,
                (
                    "Inconsistent internal state for resource %(resource)s. The inmanta resource scheduler"
                    " assumed it was still blocked, pending a successful deploy of at least one of its requires."
                    " However, all dependencies are already in a deployed state. This was expected to be impossible"
                    " and it indicates a (non-critical) bug in the inmanta resource scheduler. Please report this"
                    " bug."
                    " In the meantime, if you encounter any resources stuck in the skipped state, trigger a repair"
                    " as a workaround to force a deploy."
                ),
                resource=deploy_intent.intent.resource_id,
            )

    async def _get_last_non_deploying_state_for_dependencies(
        self, resource: ResourceIdStr
    ) -> Mapping[ResourceIdStr, const.ResourceState]:
        """
        Get resource state for every dependency of a given resource from the scheduler state.
        The state is then converted to const.ResourceState.

        Should only be called under scheduler lock.

        :param resource: The id of the resource to find the dependencies for
        """
        requires_view: Mapping[ResourceIdStr, Set[ResourceIdStr]] = self._state.requires.requires_view()
        dependencies: Set[ResourceIdStr] = requires_view.get(resource, set())
        dependencies_state = {}
        for dep_id in dependencies:
            resource_state_object: ResourceState = self._state.resource_state[dep_id]
            match resource_state_object:
                case ResourceState(compliance=Compliance.UNDEFINED):
                    dependencies_state[dep_id] = const.ResourceState.undefined
                case ResourceState(blocked=Blocked.BLOCKED):
                    dependencies_state[dep_id] = const.ResourceState.skipped_for_undefined
                case ResourceState(compliance=Compliance.HAS_UPDATE):
                    dependencies_state[dep_id] = const.ResourceState.available
                case ResourceState(last_deploy_result=DeployResult.SKIPPED):
                    dependencies_state[dep_id] = const.ResourceState.skipped
                case ResourceState(last_deploy_result=DeployResult.DEPLOYED):
                    dependencies_state[dep_id] = const.ResourceState.deployed
                case ResourceState(last_deploy_result=DeployResult.FAILED):
                    dependencies_state[dep_id] = const.ResourceState.failed
                case _:
                    raise Exception(f"Failed to parse the resource state for {dep_id}: {resource_state_object}")
        return dependencies_state

    async def get_resource_state(self) -> SchedulerStatusReport:
        """
        Check that the state of the resources in the DB corresponds
        to the internal state of the scheduler and return a SchedulerStatusReport
        object containing the internal state and any discrepancies between this
        state and that of the DB.
        """

        async def _build_discrepancy_map(
            resource_states_in_db: Mapping[ResourceIdStr, const.ResourceState],
        ) -> dict[ResourceIdStr, list[Discrepancy]]:
            """
            For each resource in the given map, compare its persisted state in the database to its
            state as it is assumed by the scheduler. Build and return a map of all detected discrepancies.

            :param resource_states_in_db: Map each resource to its ResourceState.
            :return: A dict mapping each resource to the discrepancies related to it (if any)
            """
            state_translation_table: dict[
                const.ResourceState, tuple[DeployResult | None, Blocked | None, Compliance | None]
            ] = {
                # A table to translate the old states into the new states
                # None means don't care, mostly used for values we can't derive from the old state
                const.ResourceState.unavailable: (None, Blocked.NOT_BLOCKED, Compliance.NON_COMPLIANT),
                const.ResourceState.skipped: (DeployResult.SKIPPED, None, None),
                const.ResourceState.dry: (None, None, None),  # don't care
                const.ResourceState.deployed: (DeployResult.DEPLOYED, Blocked.NOT_BLOCKED, None),
                const.ResourceState.failed: (DeployResult.FAILED, Blocked.NOT_BLOCKED, None),
                const.ResourceState.deploying: (None, Blocked.NOT_BLOCKED, None),
                const.ResourceState.available: (None, Blocked.NOT_BLOCKED, Compliance.HAS_UPDATE),
                const.ResourceState.undefined: (None, Blocked.BLOCKED, Compliance.UNDEFINED),
                const.ResourceState.skipped_for_undefined: (None, Blocked.BLOCKED, None),
            }

            discrepancy_map: dict[ResourceIdStr, list[Discrepancy]] = {}

            # Resources only present in the DB but missing from the scheduler
            only_in_db = resource_states_in_db.keys() - self._state.resource_state.keys()
            for rid in only_in_db:
                discrepancy_map[rid] = [
                    Discrepancy(
                        rid=rid,
                        field=None,
                        expected=(
                            f"Resource is present in the DB (model version {latest_version}). "
                            "It is expected in the scheduler's state map."
                        ),
                        actual="Resource is missing from the scheduler's state map.",
                    )
                ]

            # Resources only present in the scheduler but missing from the DB
            only_in_scheduler = self._state.resource_state.keys() - resource_states_in_db.keys()
            for rid in only_in_scheduler:
                discrepancy_map[rid] = [
                    Discrepancy(
                        rid=rid,
                        field=None,
                        expected=(
                            f"Resource is not present in the DB (model version {latest_version}). "
                            "It shouldn't be in the scheduler's state map."
                        ),
                        actual="Resource is present in the scheduler's state map.",
                    )
                ]

            # Keep track of the number of iterations to regularly pass control back to the io loop
            i: int = 0
            # For resources in both the DB and the scheduler, check for discrepancies in state
            for rid in resource_states_in_db.keys() & self._state.resource_state.keys():
                resource_discrepancies: list[Discrepancy] = []

                db_resource_status = resource_states_in_db[rid]
                db_deploy_result, db_blocked_status, db_compliance_status = state_translation_table[db_resource_status]

                scheduler_resource_state: ResourceState = self._state.resource_state[rid]
                if db_deploy_result:
                    if scheduler_resource_state.last_deploy_result != db_deploy_result:
                        resource_discrepancies.append(
                            Discrepancy(
                                rid=rid,
                                field="last_deploy_result",
                                expected=db_deploy_result,
                                actual=scheduler_resource_state.last_deploy_result,
                            )
                        )
                if db_blocked_status:
                    if scheduler_resource_state.blocked != db_blocked_status:
                        resource_discrepancies.append(
                            Discrepancy(
                                rid=rid,
                                field="blocked_status",
                                expected=db_blocked_status,
                                actual=scheduler_resource_state.blocked,
                            )
                        )
                if db_compliance_status:
                    if scheduler_resource_state.compliance != db_compliance_status:
                        resource_discrepancies.append(
                            Discrepancy(
                                rid=rid,
                                field="compliance_status",
                                expected=db_compliance_status,
                                actual=scheduler_resource_state.compliance,
                            )
                        )

                if resource_discrepancies:
                    discrepancy_map[rid] = resource_discrepancies

                # Pass back control to the io loop every 100 iterations to not
                # block scheduler operations in case we have a lot of resources.
                if i >= NB_ITERATIONS_PASS_IO_LOOP:
                    await asyncio.sleep(0)
                    i = 0
                i += 1

            return discrepancy_map

        def report_model_version_mismatch(db_version: int | None) -> SchedulerStatusReport:
            LOGGER.info(
                "Cannot compare scheduler status and database status because of "
                "model version mismatch (%s and %s respectively). "
                "The scheduler might not have processed this new version yet.",
                self._state.version,
                db_version,
            )
            return SchedulerStatusReport(
                scheduler_state={},
                db_state={},
                resource_states={},
                discrepancies=[
                    Discrepancy(
                        rid=None,
                        field="model_version",
                        expected=str(db_version),
                        actual=str(self._state.version),
                    )
                ],
            )

        latest_version: int | None
        async with self._scheduler_lock:
            async with self.state_update_manager.get_connection() as connection:
                try:
                    latest_model: ModelVersion = await self._get_single_model_version_from_db(connection=connection)
                except KeyError:
                    return SchedulerStatusReport(scheduler_state={}, db_state={}, resource_states={}, discrepancies={})
                if latest_model.version != self._state.version:
                    return report_model_version_mismatch(latest_model.version)

                resource_states_in_db: Mapping[ResourceIdStr, const.ResourceState]
                latest_version, resource_states_in_db = await data.Resource.get_resource_states_latest_version(
                    env=self.environment, connection=connection
                )
                if latest_version != self._state.version:
                    return report_model_version_mismatch(latest_version)

            discrepancy_map = await _build_discrepancy_map(resource_states_in_db=resource_states_in_db)
            return SchedulerStatusReport(
                scheduler_state=self._state.resource_state,
                db_state=latest_model.resources,
                resource_states=resource_states_in_db,
                discrepancies=discrepancy_map,
            )

    async def suspend_deployments(self, reason: str) -> None:
        """
        Suspend all agent operations. All other scheduler functionality remains active.
        """
        LOGGER.info("Suspending all deployment operations: %s", reason)
        self._deployment_suspended = True
        await asyncio.gather(*[worker.stop() for worker in self._workers.values()])
        self._work.add_poison_pill_to_agent_queues(reason=reason)
        await asyncio.gather(*[worker.join() for worker in self._workers.values()])

    async def resume_deployments(self) -> None:
        """
        Resume all agent operations.
        """
        LOGGER.info("Resuming all deployment operations.")
        self._deployment_suspended = False
        await self.refresh_all_agent_states_from_db()



"""
Copyright 2021 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import os
import pathlib

import pytest

import inmanta.util
import packaging.requirements
from inmanta.file_parser import RequirementsTxtParser


def test_requirements_txt_parser(tmpdir) -> None:
    content = """
        test==1.2.3

        # A comment
        other-dep~=2.0.0
        third-dep<5.0.0 # another comment
        split\
ted\
dep
        Capital
    """

    requirements_txt_file = os.path.join(tmpdir, "requirements.txt")
    with open(requirements_txt_file, "w", encoding="utf-8") as fd:
        fd.write(content)

    expected_requirements = ["test==1.2.3", "other-dep~=2.0.0", "third-dep<5.0.0", "splitteddep", "Capital"]
    requirements: list[inmanta.util.CanonicalRequirement] = RequirementsTxtParser().parse(requirements_txt_file)
    assert requirements == inmanta.util.parse_requirements(expected_requirements)
    requirements_as_str = RequirementsTxtParser.parse_requirements_as_strs(requirements_txt_file)
    assert requirements_as_str == expected_requirements

    parsed_canonical_requirements_from_file = inmanta.util.parse_requirements_from_file(pathlib.Path(requirements_txt_file))
    assert parsed_canonical_requirements_from_file == requirements

    problematic_requirements = [
        "test==1.2.3",
        "other-dep~=2.0.0",
        "third-dep<5.0.0 # another comment",
        "splitteddep",
        "Capital",
    ]

    parsed_canonical_requirements = inmanta.util.parse_requirements(expected_requirements)
    assert parsed_canonical_requirements == requirements

    with pytest.raises(Exception) as e:
        inmanta.util.parse_requirements(problematic_requirements)
    assert "Expected end or semicolon (after version specifier)\n    third-dep<5.0.0 # another comment\n" in str(e.value)

    new_content = RequirementsTxtParser.get_content_with_dep_removed(requirements_txt_file, remove_dep_on_pkg="test")
    expected_content = """

        # A comment
        other-dep~=2.0.0
        third-dep<5.0.0 # another comment
        splitteddep
        Capital
    """
    assert new_content == expected_content
    new_content = RequirementsTxtParser.get_content_with_dep_removed(requirements_txt_file, remove_dep_on_pkg="third-dep")
    assert (
        new_content
        == """
        test==1.2.3

        # A comment
        other-dep~=2.0.0
        splitteddep
        Capital
    """
    )
    new_content = RequirementsTxtParser.get_content_with_dep_removed(requirements_txt_file, remove_dep_on_pkg="splitteddep")
    assert (
        new_content
        == """
        test==1.2.3

        # A comment
        other-dep~=2.0.0
        third-dep<5.0.0 # another comment
        Capital
    """
    )
    new_content = RequirementsTxtParser.get_content_with_dep_removed(requirements_txt_file, remove_dep_on_pkg="Capital")
    assert (
        new_content
        == """
        test==1.2.3

        # A comment
        other-dep~=2.0.0
        third-dep<5.0.0 # another comment
        splitteddep
    """
    )


@pytest.mark.parametrize(
    "iteration",
    [
        ("", True),
        ("#", True),
        ("   # ", True),
        ("#this is a comment", True),
        ("test==1.2.3", False),
        ("other-dep~=2.0.0", False),
    ],
)
def test_canonical_requirement(iteration) -> None:
    """
    Ensure that empty name requirements are not allowed in `Requirement`
    """
    name, should_fail = iteration
    if should_fail:
        with pytest.raises(packaging.requirements.InvalidRequirement):
            inmanta.util.parse_requirement(requirement=name)
    else:
        inmanta.util.parse_requirement(requirement=name)


@pytest.mark.parametrize(
    "iteration",
    [
        ("", ""),
        ("#", "#"),
        ("   # ", "#"),
        ("#this is a comment", "#this is a comment"),
        ("test==1.2.3", "test==1.2.3"),
        ("other-dep~=2.0.0", "other-dep~=2.0.0"),
        ("test==1.2.3  # a command", "test==1.2.3"),
        ("other-dep #~=2.0.0", "other-dep"),
        ("other-dep#~=2.0.0", "other-dep#~=2.0.0"),
    ],
)
def test_drop_comment_part(iteration) -> None:
    """
    Ensure that empty name requirements are not allowed in `Requirement`
    """
    value, expected_value = iteration
    current_value = inmanta.util.remove_comment_part_from_specifier(value)
    assert current_value == expected_value



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import contextvars
import dataclasses
import enum
from collections.abc import Iterator, Mapping, Sequence
from copy import copy
from dataclasses import is_dataclass
from types import TracebackType
from typing import TYPE_CHECKING, Callable, ContextManager, Optional, Self, Union

# Keep UnsetException, UnknownException and AttributeNotFound in place for backward compat with <iso8
from inmanta import references
from inmanta.ast import UnsetException  # noqa F401
from inmanta.ast import AttributeNotFound, Location, NotFoundException, RuntimeException, UnexpectedReference, UnknownException
from inmanta.execute.util import NoneValue, Unknown
from inmanta.stable_api import stable_api
from inmanta.types import PrimitiveTypes
from inmanta.util import JSONSerializable

if TYPE_CHECKING:
    from inmanta.ast.entity import Entity
    from inmanta.ast.type import Type as inm_Type
    from inmanta.execute.runtime import Instance, QueueScheduler, Resolver

    TypeResolver = Callable[[type[object]], inm_Type]
else:
    TypeResolver = object


@dataclasses.dataclass(frozen=True, slots=True)
class DynamicUnwrapContext:
    """A set of context information that allows dynamic proxy unwrapping to construct instances"""

    resolver: "Resolver"
    queue: "QueueScheduler"
    location: Location
    # this last one is purely there to prevent import loops
    type_resolver: TypeResolver


class ProxyMode(enum.Enum):
    """
    The mode to proxy values for. Export mode is more lax than plugin mode when it comes to references.
    """

    PLUGIN = enum.auto()
    EXPORT = enum.auto()


global_proxy_mode: contextvars.ContextVar[ProxyMode] = contextvars.ContextVar("global_proxy_mode", default=ProxyMode.PLUGIN)
"""
This variable controls the behavior of all proxy objects

It is global variable for performance reasons. Dynamic proxies are extremely performance sensitive because the are use a lot
"""


class ExportContext(ContextManager[None]):

    def __enter__(self) -> None:
        global_proxy_mode.set(ProxyMode.EXPORT)

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, traceback: TracebackType | None
    ) -> None:
        global_proxy_mode.set(ProxyMode.PLUGIN)


exportcontext = ExportContext()


class ProxyContext:
    """
    Context for creating proxy objects. Declares whether the object has already passed certain validation or whether certain
    special values are expected in the object's values (attributes, list elements, ...).

    E.g. a top-level plugin argument will have been validated at the boundary, and an undeclared reference would have been
    rejected by that validation. In contrast, an opaque DSL instance's attributes have had no boundary validation, and may
    contain references that are not (can not be) declared. Since we require reference support to always be explicit, we have
    to reject reference values in such cases.

    The global_proxy_mode determines the overall behavior:
    1. when in export mode, references are always allowed
    2. when in plugin mode, reference access is controlled by allow_reference_values and validated


    :param validated: True iff the object to proxy has been validated at the plugin boundary.
    :param allow_reference_values: Allow references for values accessed through this proxy. Either because they have been
        declared and validated, or because explicitly requested. Defaults to allow references iff the object has been
        validated. Kept as a separate field to allow overrides for a single proxy object while maintaining behavior for
        nested proxies. This value is not passed on to children.
    :param path: The path, within the plugin's argument namespace, where this object lives. Should be compatible with
        composition at the tail, e.g. `list_arg[0].attr`.
    """

    __slots__ = ("path", "validated", "allow_reference_values")

    def __init__(self, *, path: str, validated: bool = True, allow_reference_values: Optional[bool] = None) -> None:
        self.path = path
        self.validated = validated
        self.allow_reference_values = allow_reference_values

    def nested(self, *, relative_path: str) -> "ProxyContext":
        """
        Returns a context object for values nested one level deeper than the current context.

        :param relative_path: The path of new object, relative to the current context.
        """
        return ProxyContext(
            path=self.path + relative_path,
            validated=self.validated,
            # we're proxying elements one level deeper than the current context
            # => reset allow_reference_values to default behavior
            allow_reference_values=None,
        )

    def allow_references(self) -> "ProxyContext":
        return ProxyContext(
            path=self.path,
            validated=self.validated,
            allow_reference_values=True,
        )

    def should_allow_references(self) -> bool:
        if global_proxy_mode.get() is ProxyMode.EXPORT:
            return True
        return self.allow_reference_values if self.allow_reference_values is not None else self.validated


# this is here to avoid import loops
# It would be nicer to have it as class method on entity, but that would cause proxy to import the entire compiler
def get_inmanta_type_for_dataclass(for_type: type[object]) -> "Entity | None":
    if hasattr(for_type, "_paired_inmanta_entity"):
        return for_type._paired_inmanta_entity
    return None


@stable_api
class DynamicProxy:
    """
    This class wraps an object and makes sure that a model is never modified
    by native code.
    """

    def __init__(self, instance: "Instance", *, context: Optional[ProxyContext] = None) -> None:
        """
        :param instance: The object to proxy.
        :param context: The context this object lives in.
        """
        # str() is expensive. And also not helpful for e.g. dicts.
        # => Fall back to object.__repr__ for externally created objects (internal calls should all pass a context)
        context = context if context is not None else ProxyContext(path=object.__repr__(instance))
        object.__setattr__(self, "__instance", instance)
        object.__setattr__(self, "__context", context)

    def _get_instance(self) -> "Instance":
        return object.__getattribute__(self, "__instance")

    def _get_context(self) -> ProxyContext:
        return object.__getattribute__(self, "__context")

    def _allow_references(self: Self) -> Self:
        """
        Returns a copy of this proxy object that allows access to its elements even if they are references.

        Allows references for a single object, not nested.
        """
        # don't just call constructor for backwards compatibility: some children outside of core might not have context arg
        new: Self = copy(self)
        object.__setattr__(new, "__context", self._get_context().allow_references())
        return new

    @classmethod
    def unwrap(cls, item: object, *, dynamic_context: DynamicUnwrapContext | None = None) -> object:
        """
        Converts a value from the plugin domain to the internal domain.

        :param dynamic_context: a type resolver context. When passed in, dataclasses are converted as well.
        """
        if item is None:
            return NoneValue()

        if isinstance(item, DynamicProxy):
            return item._get_instance()

        if isinstance(item, list):
            return [cls.unwrap(x, dynamic_context=dynamic_context) for x in item]

        if isinstance(item, dict):

            def recurse_dict_item(key_value: tuple[object, object]) -> tuple[object, object]:
                (key, value) = key_value
                if not isinstance(key, str):
                    raise RuntimeException(
                        None, f"dict keys should be strings, got {key!r} of type {type(key)} with dict value {value!r}"
                    )
                return (key, cls.unwrap(value, dynamic_context=dynamic_context))

            return dict(map(recurse_dict_item, item.items()))

        if isinstance(item, references.Reference):
            ref_type = item.get_reference_type()
            if ref_type is None:
                raise RuntimeException(
                    None,
                    f"Could not determine the reference type of {item}, "
                    f"make sure the reference extends Reference[concrete_type] or override `get_reference_type`",
                )

            if dataclasses.is_dataclass(ref_type):
                if dynamic_context is None:
                    raise RuntimeException(
                        None,
                        f"{item} is a dataclass of type {ref_type}. "
                        "It can only be converted to an inmanta entity at the plugin boundary",
                    )
                dataclass_ref_type = dynamic_context.type_resolver(ref_type)
                item._model_type = dataclass_ref_type
                # Can not be typed correctly due to import loops
                return dataclass_ref_type.from_python(
                    item, dynamic_context.resolver, dynamic_context.queue, dynamic_context.location
                )
            else:
                if item._model_type is None:
                    if dynamic_context is None:
                        raise RuntimeException(
                            None,
                            f"{item} is a reference of type {ref_type}. "
                            "It can only be typed at the plugin boundary or "
                            "by explicitly setting `_model_type` to the relevant inmanta type",
                        )
                    reference_type = dynamic_context.type_resolver(ref_type)
                    item._model_type = reference_type
                return item

        if is_dataclass(item) and not isinstance(item, type):
            # dataclass instance
            dataclass_type = get_inmanta_type_for_dataclass(type(item))
            if dataclass_type is not None:
                if dynamic_context is not None:
                    return dataclass_type.from_python(
                        item, dynamic_context.resolver, dynamic_context.queue, dynamic_context.location
                    )
                else:
                    raise RuntimeException(
                        None,
                        f"{item} is a dataclass of type {dataclass_type.get_full_name()}. "
                        "It can only be converted to an inmanta entity at the plugin boundary",
                    )

        return item

    @classmethod
    def return_value(
        cls,
        value: object,
        *,
        context: Optional[ProxyContext] = None,  # optional for backwards compatibility
    ) -> Union[None, str, tuple[object, ...], int, float, bool, "DynamicProxy", "references.Reference"]:
        """
        Converts a value from the internal domain to the plugin domain.

        :param context: The context in which the given object lives. If None, assumes that the value has already passed
            validation. When None, the object's string representation, rather than its name is used for error reporting.
        """

        if value is None:
            return None

        if isinstance(value, NoneValue):
            return None

        if isinstance(value, Unknown):
            raise UnknownException(value)

        if isinstance(value, (str, tuple, int, float, bool)):
            return copy(value)

        if isinstance(value, references.Reference):
            # if a reference gets here, it has been validated, and we want to represent it as a reference, not a proxy
            return value

        if isinstance(value, DynamicProxy):
            return value

        # str() is expensive. And also not helpful for e.g. dicts.
        # => Fall back to object.__repr__ for externally created objects (internal calls should all pass a context)
        context = context if context is not None else ProxyContext(path=object.__repr__(value))

        if isinstance(value, dict):
            return DictProxy(value, context=context)

        if hasattr(value, "__len__"):
            return SequenceProxy(value, context=context)

        if hasattr(value, "__call__"):
            return CallProxy(value, context=context)

        return DynamicProxy(
            value,
            # DSL instances are a black box as far as boundary validation is concerned
            # => from here on out, consider the object not validated, except during export
            context=(
                ProxyContext(
                    path=context.path,
                    validated=False,
                    allow_reference_values=None,
                )
                if context.validated
                else context
            ),
        )

    def _return_value(self, value: object, *, relative_path: str) -> object:
        """
        Return a value that was accessed through this proxy object. Validates for undeclared references and propagates context
        appropriately.
        """
        context: ProxyContext = self._get_context()
        value_context: ProxyContext = context.nested(relative_path=relative_path)

        if isinstance(value, references.Reference) and not context.should_allow_references():
            # Non-dataclass entities can not be explicit about reference support.
            # The Python domain is a black box. We don't want to transparently pass unexpected values in there.
            # => don't allow references in attributes. Can be explicitly allowed via allow_reference_attributes() wrapper
            raise UnexpectedReference(
                reference=value,
                message=(
                    "Encountered unexpected reference value during plugin execution. Plugins are only allowed to access"
                    " reference values when declared explicitly. Either use a dataclass entity that supports references (e.g."
                    " `int | Reference[int]` attribute annotation), or explicitly allow references on attribute access with the"
                    " `inmanta.plugins.allow_reference_values()` wrapper."
                    f" Encountered at {value_context.path} (= `{value!r}`)."
                ),
            )

        return DynamicProxy.return_value(value, context=value_context)

    def __getattr__(self, attribute: str):
        instance = self._get_instance()

        try:
            value = instance.get_attribute(attribute).get_value()
        except NotFoundException as e:
            # allow for hasattr(proxy, "some_attr")
            raise AttributeNotFound(e.stmt, e.name)

        return self._return_value(value, relative_path=f".{attribute}")

    def __setattr__(self, attribute: str, value: object) -> None:
        raise Exception("Readonly object")

    def _type(self) -> "Entity":
        """
        Return the type of the proxied instance
        """
        return self._get_instance().type

    def is_unknown(self) -> bool:
        """
        Return true if this value is unknown and cannot be determined
        during this compilation run
        """
        if isinstance(self._get_instance(), Unknown):
            return True
        return False

    def __hash__(self) -> int:
        return hash(self._get_instance())

    def __eq__(self, other: object) -> bool:
        if hasattr(other, "_get_instance"):
            other = other._get_instance()

        return self._get_instance() == other

    def __lt__(self, other: object) -> bool:
        if hasattr(other, "_get_instance"):
            other = other._get_instance()

        return self._get_instance() < other

    def __repr__(self) -> str:
        return "@%s" % repr(self._get_instance())


class SequenceProxy(DynamicProxy, JSONSerializable):
    def __init__(self, iterator: Sequence[object], *, context: Optional[ProxyContext] = None) -> None:
        DynamicProxy.__init__(self, iterator, context=context)

    def __getitem__(self, key: int) -> object:
        instance = self._get_instance()
        if isinstance(key, str):
            raise RuntimeException(self, f"can not get a attribute {key}, {self._get_instance()} is a list")

        return self._return_value(instance[key], relative_path=f"[{key!r}]")

    def __iter__(self):
        return IteratorProxy(iter(self._get_instance()), context=self._get_context(), sequence=True)

    def __len__(self) -> int:
        return len(self._get_instance())

    def json_serialization_step(self) -> list[PrimitiveTypes]:
        # Ensure proper unwrapping by using __getitem__
        return [i for i in self]


class DictProxy(DynamicProxy, Mapping[str, object], JSONSerializable):
    def __init__(self, mydict: dict[str, object], *, context: Optional[ProxyContext] = None) -> None:
        DynamicProxy.__init__(self, mydict, context=context)

    def __getitem__(self, key):
        instance = self._get_instance()
        if not isinstance(key, str):
            raise RuntimeException(self, f"Expected string key, but got {key}, {self._get_instance()} is a dict")

        return self._return_value(instance[key], relative_path=f"[{key!r}]")

    def __iter__(self):
        return IteratorProxy(iter(self._get_instance()), context=self._get_context(), sequence=False)

    def __len__(self) -> int:
        return len(self._get_instance())

    def json_serialization_step(self) -> dict[str, PrimitiveTypes]:
        # Ensure proper unwrapping by using __getitem__
        return {k: v for k, v in self.items()}


class IteratorProxy(DynamicProxy, Iterator[object]):
    """
    Proxy an iterator call.

    A custom proxy allows us to continue iteration after exceptions are raised and caught by the caller, i.e. special exception
    types like UnknownException.
    """

    def __init__(self, iterator: Iterator[object], *, context: Optional[ProxyContext] = None, sequence: bool = False) -> None:
        """
        :param sequence: True iff the given iterator represents a sequence, i.e. the index is meaningful for error reporting.
        """
        DynamicProxy.__init__(self, enumerate(iterator), context=context)
        object.__setattr__(self, "__sequence", sequence)

    def _is_sequence(self) -> bool:
        return object.__getattribute__(self, "__sequence")

    def __iter__(self) -> Iterator[object]:
        return self

    def __next__(self) -> object:
        enumerator = self._get_instance()
        i, v = next(enumerator)
        return self._return_value(
            v,
            # if it's not a sequence, pointing to the object itself is the best we can do
            relative_path=f"[{i}]" if self._is_sequence() else "",
        )


class CallProxy(DynamicProxy):
    """
    Proxy a value that implements a __call__ function
    """

    def __init__(self, instance: Callable[..., object], *, context: Optional[ProxyContext] = None) -> None:
        DynamicProxy.__init__(self, instance, context=context)

    def __call__(self, *args, **kwargs):
        instance = self._get_instance()

        return instance(*args, **kwargs)



"""
Copyright 2018 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import logging
import os
import re
import signal
import subprocess
import sys
import time
import typing
import uuid
from subprocess import TimeoutExpired
from threading import Timer

import pytest

import inmanta.util
from inmanta import const
from inmanta.app import CompileSummaryReporter

LOGGER = logging.getLogger(__name__)


def get_command(
    tmp_dir,
    stdout_log_level=None,
    log_file=None,
    log_level_log_file=None,
    timed=False,
    dbport=None,
    dbname="inmanta",
    dbhost=None,
    dbuser=None,
    dbpass=None,
    config_dir=None,
    server_extensions=[],
    version=False,
    command: str = "server",
):
    """Build an argument string for subprocess to run the orchestrator inmanta.app entrypoint"""
    root_dir = tmp_dir.mkdir("root").strpath
    log_dir = os.path.join(root_dir, "log")
    state_dir = os.path.join(root_dir, "data")
    for directory in [log_dir, state_dir]:
        os.mkdir(directory)
    config_file = os.path.join(root_dir, "inmanta.cfg")

    if dbport is not None:
        port = dbport
    else:
        port = inmanta.util.get_free_tcp_port()

    with open(config_file, "w+", encoding="utf-8") as f:
        f.write("[config]\n")
        f.write("log-dir=" + log_dir + "\n")
        f.write("state-dir=" + state_dir + "\n")
        f.write("[database]\n")
        f.write("port=" + str(port) + "\n")
        f.write("name=" + dbname + "\n")
        if dbhost:
            f.write(f"host={dbhost}\n")
        if dbuser:
            f.write(f"username={dbuser}\n")
        if dbpass:
            f.write(f"password={dbpass}\n")
        f.write("[server]\n")
        f.write(f"enabled_extensions={', '.join(server_extensions)}\n")

    args = [sys.executable, "-m", "inmanta.app"]
    if stdout_log_level:
        args.append("-" + "v" * stdout_log_level)
    if log_file:
        log_file = os.path.join(log_dir, log_file)
        args += ["--log-file", log_file]
    if log_file and log_level_log_file:
        args += ["--log-file-level", str(log_level_log_file)]
    if timed:
        args += ["--timed-logs"]
    if config_dir:
        args += ["--config-dir", config_dir]
    if version:
        args += ["--version"]
    args += ["-c", config_file, command]
    return (args, log_dir)


def do_run(args: list[str], env: typing.Optional[dict[str, str]] = None, cwd: typing.Optional[str] = None) -> subprocess.Popen:
    if env is None:
        env = {}
    env_display = "parent process env vars"
    if env:
        env_display += " and the following extra env vars %s" % env
    LOGGER.info("Running %s with %s and cwd %s.", args, env_display, cwd)
    baseenv = os.environ.copy()
    baseenv.update(env)
    process = subprocess.Popen(args, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=baseenv)
    return process


def convert_to_ascii(text):
    return [line for line in text.decode("ascii").split("\n") if line != ""]


def do_kill(process: subprocess.Popen, killtime: int = 3, termtime: int = 2) -> tuple[str, str, int]:
    """Terminate the process after termtime and kill it after killtime"""

    def do_and_log(func, msg):
        def w():
            LOGGER.warning(msg)
            func()

        return w

    t1 = Timer(killtime, do_and_log(process.kill, f"killed process after {killtime}s"))
    t2 = Timer(termtime, do_and_log(process.terminate, f"terminated process after {termtime}s"))
    t1.start()
    t2.start()
    start = time.time()
    out, err = process.communicate()
    end = time.time()

    LOGGER.info("Run took at least %f s", (end - start))
    t1.cancel()
    t2.cancel()

    stdout = convert_to_ascii(out)
    stderr = convert_to_ascii(err)
    return (stdout, stderr, process.returncode)


def run_without_tty(
    args: list[str],
    env: typing.Optional[dict[str, str]] = None,
    killtime: int = 15,
    termtime: int = 10,
    *,
    treat_warnings_as_errors: bool = False,
) -> tuple[str, str, int]:
    """Run the given command without a tty"""
    if treat_warnings_as_errors:
        if env is None:
            env = {}
        env["PYTHONWARNINGS"] = "error"

    process = do_run(args, env)
    return do_kill(process, killtime, termtime)


def run_with_tty(args, killtime=4, termtime=3, *, treat_warnings_as_errors: bool = False):
    """Could not get code for actual tty to run stable in docker, so we are faking it"""
    env = {const.ENVIRON_FORCE_TTY: "true"}
    return run_without_tty(
        args, env=env, killtime=killtime, termtime=termtime, treat_warnings_as_errors=treat_warnings_as_errors
    )


def is_colorama_package_available():
    try:
        import colorama  # noqa: F401
    except ModuleNotFoundError:
        return False
    return True


def test_verify_that_colorama_package_is_not_present():
    """
    The colorama package turns the colored characters in TTY-based terminal into uncolored characters.
    As such, this package should not be present.
    """
    assert not is_colorama_package_available()


# Log lines emitted by the --version command of inmanta.app on the cli or a log file (with logger between level and message)
INFO_MSG = r"INFO\s+[a-z\.]*\s+Discovered extensions:"
DEBUG_MSG = r"DEBUG\s+[a-z\.]*\s+Using selector: EpollSelector"
INFO_MSG_TTY = r"INFO\s+[a-z\.]*\s+\x1b\[0m\x1b\[34mDiscovered extensions:"
DEBUG_MSG_TTY = r"DEBUG\s+[a-z\.]*\s+\x1b\[0m\x1b\[34mUsing selector: EpollSelector"

# The number of v arguments provided to get this level of logs
LEVEL_WARNING = 1
LEVEL_INFO = 2
LEVEL_DEBUG = 3


@pytest.mark.parametrize_any(
    "log_level, timed, with_tty, regexes_required_lines, regexes_forbidden_lines",
    [
        (LEVEL_DEBUG, False, False, [INFO_MSG, DEBUG_MSG], []),
        (LEVEL_INFO, False, False, [INFO_MSG], [DEBUG_MSG]),
        (LEVEL_DEBUG, False, True, [INFO_MSG_TTY, DEBUG_MSG_TTY], []),
        (LEVEL_INFO, False, True, [INFO_MSG_TTY], [DEBUG_MSG_TTY]),
        (LEVEL_DEBUG, True, False, [INFO_MSG, DEBUG_MSG], []),
        (LEVEL_INFO, True, False, [INFO_MSG], [DEBUG_MSG]),
        (LEVEL_DEBUG, True, True, [INFO_MSG_TTY, DEBUG_MSG_TTY], []),
        (LEVEL_INFO, True, True, [INFO_MSG_TTY], [DEBUG_MSG_TTY]),
    ],
)
@pytest.mark.timeout(20)
def test_no_log_file_set(tmpdir, log_level, timed, with_tty, regexes_required_lines, regexes_forbidden_lines):
    """Test if"""
    if is_colorama_package_available() and with_tty:
        pytest.skip("Colorama is present")

    (args, log_dir) = get_command(tmpdir, stdout_log_level=log_level, timed=timed, command="--version")
    if with_tty:
        (stdout, _, _) = run_with_tty(args)
    else:
        (stdout, _, _) = run_without_tty(args)
    log_file = "server.log"
    assert log_file not in os.listdir(log_dir)
    assert len(stdout) != 0
    check_logs(stdout, regexes_required_lines, regexes_forbidden_lines, timed)


@pytest.mark.parametrize_any(
    "log_level,with_tty, regexes_required_lines, regexes_forbidden_lines",
    [
        (LEVEL_DEBUG, False, [INFO_MSG, DEBUG_MSG], []),
        ("DEBUG", False, [INFO_MSG, DEBUG_MSG], []),
        (LEVEL_INFO, False, [INFO_MSG], [DEBUG_MSG]),
        ("INFO", False, [INFO_MSG], [DEBUG_MSG]),
        (LEVEL_DEBUG, True, [INFO_MSG, DEBUG_MSG], []),
        (LEVEL_INFO, True, [INFO_MSG], [DEBUG_MSG]),
    ],
)
@pytest.mark.timeout(60)
def test_log_file_set(tmpdir, log_level, with_tty, regexes_required_lines, regexes_forbidden_lines):
    """Check if lines are logged correctly to file and those lines are not emited on the commandline"""
    if is_colorama_package_available() and with_tty:
        pytest.skip("Colorama is present")

    log_file = "server.log"
    (args, log_dir) = get_command(tmpdir, log_file=log_file, log_level_log_file=log_level, command="--version")
    if with_tty:
        (stdout, _, _) = run_with_tty(args)
    else:
        (stdout, _, _) = run_without_tty(args)
    assert log_file in os.listdir(log_dir)
    log_file = os.path.join(log_dir, log_file)
    with open(log_file) as f:
        log_lines = f.readlines()

    check_logs(log_lines, regexes_required_lines, regexes_forbidden_lines, timed=False)
    # Check if the message appears in the logs, it is not a full line match so timing is not relevant
    check_logs(stdout, [], regexes_required_lines, timed=False)


@pytest.mark.parametrize_any(
    "with_tty, regexes_required_lines",
    [
        (False, [r"Inmanta Service Orchestrator", r"Compiler version: ", r"Extensions:", r"\s*\* core:"]),
        (True, [r"Inmanta Service Orchestrator", r"Compiler version: ", r"Extensions:", r"\s*\* core:"]),
    ],
)
@pytest.mark.timeout(60)
def test_version_command(tmpdir, with_tty: bool, regexes_required_lines: list[str]):
    """
    Basic smoke test to check that no warnings show up when running 'inmanta --version'
    """
    if is_colorama_package_available() and with_tty:
        pytest.skip("Colorama is present")

    (args, log_dir) = get_command(tmpdir, command="--version")
    if with_tty:
        (stdout, stderr, return_code) = run_with_tty(args, treat_warnings_as_errors=True)
    else:
        (stdout, stderr, return_code) = run_without_tty(args, treat_warnings_as_errors=True)

    if return_code != 0 or stderr:
        e = Exception("An unexpected error occurred or warnings were logged while running 'inmanta --version'")
        if stderr:
            e.add_note("\n".join(stderr))
        if stdout:
            e.add_note("\n".join(stdout))
        raise e
    # Check if the message appears in the logs, it is not a full line match so timing is not relevant
    check_logs(stdout, regexes_required_lines, regexes_forbidden_lines=[], timed=False)


@pytest.mark.parametrize_any(
    "log_level, regexes_required_lines, regexes_forbidden_lines",
    [
        (LEVEL_DEBUG, [INFO_MSG, DEBUG_MSG], []),
        (LEVEL_INFO, [INFO_MSG], [DEBUG_MSG]),
        (LEVEL_WARNING, [], [INFO_MSG, DEBUG_MSG]),
    ],
)
@pytest.mark.timeout(60)
def test_log_stdout_log_level(log_level, regexes_required_lines, regexes_forbidden_lines):
    """Check if the inmanta command prints out the correct logs depending on the amount of provided -v flags on the CLI"""
    args = [sys.executable, "-m", "inmanta.app", "-" + "v" * log_level, "--version"]
    logging.getLogger(__name__).info("Starting inmanta: %s", args)
    (stdout, err, _) = run_without_tty(args)
    check_logs(stdout, regexes_required_lines, regexes_forbidden_lines, timed=False)


def check_logs(log_lines, regexes_required_lines, regexes_forbidden_lines, timed):
    if not log_lines:
        print("No lines logged")

    for line in log_lines:
        print(line)

    for regex in regexes_required_lines:
        if not any(re.search(regex, line) for line in log_lines):
            pytest.fail(f"Required pattern was not found in log lines: {regex}")

        if timed and not any(
            re.match(r"[\d]{4}\-[\d]{2}\-[\d]{2} [\d]{2}\:[\d]{2}\:[\d]{2}\,[\d]{3}", line) for line in log_lines
        ):
            pytest.fail("Timed log should start with timestamp.")

    for regex in regexes_forbidden_lines:
        if any(re.search(regex, line) for line in log_lines):
            pytest.fail(f"Forbidden pattern found in log lines: {regex}")


def test_check_shutdown():
    process = do_run([sys.executable, os.path.join(os.path.dirname(__file__), "miniapp.py")])
    # wait for handler to be in place
    try:
        process.communicate(timeout=2)
    except TimeoutExpired:
        pass
    process.send_signal(signal.SIGUSR1)
    out, err, code = do_kill(process, killtime=3, termtime=1)
    print(out, err)
    assert code == 0
    assert "----- Thread Dump ----" in out
    assert "STOP" in out
    assert "SHUTDOWN COMPLETE" in out


def test_check_bad_shutdown():
    print([sys.executable, os.path.join(os.path.dirname(__file__), "miniapp.py"), "bad"])
    process = do_run([sys.executable, os.path.join(os.path.dirname(__file__), "miniapp.py"), "bad"])
    out, err, code = do_kill(process, killtime=5, termtime=2)
    print(out, err)
    assert code == 3
    assert "----- Thread Dump ----" in out
    assert "STOP" not in out
    assert "SHUTDOWN COMPLETE" not in out


def test_startup_failure(tmpdir, postgres_db, database_name):
    (args, log_dir) = get_command(
        tmpdir,
        dbport=postgres_db.port,
        dbname=database_name,
        dbhost=postgres_db.host,
        dbuser=postgres_db.user,
        dbpass=postgres_db.password,
        server_extensions=["badplugin"],
    )
    pp = ":".join(sys.path)
    # Add a bad module
    extrapath = os.path.join(os.path.dirname(__file__), "data", "bad_module_path")
    (stdout, stderr, code) = run_without_tty(args, env={"PYTHONPATH": pp + ":" + extrapath}, killtime=15, termtime=10)
    assert "inmanta                  ERROR   Server setup failed" in stdout
    assert (
        "                                 " + "inmanta.server.protocol.SliceStartupException: Slice badplugin.badslice "
        "failed to start because: Too bad, this plugin is broken"
    ) in stdout
    assert code == 4


@pytest.mark.parametrize("cache_cf_files", [True, False])
def test_compiler_exception_output(snippetcompiler, cache_cf_files):
    """
    This test case is also used to test the caching (issue 3838)
    Since this is a basic smoke test for argument parsing, no assertion
    about the caching is done here.
    """
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    int attr
end

implement Test using std::none

o = Test(attr="1234")
        """,
        autostd=True,
    )
    cwd = snippetcompiler.project_dir if cache_cf_files else "."

    output = (
        f"""Could not set attribute `attr` on instance `__config__::Test (instantiated at {cwd}/main.cf:8)` """
        f"""(reported in Construct(Test) ({cwd}/main.cf:8))
caused by:
  Invalid value '1234', expected int (reported in Construct(Test) ({cwd}/main.cf:8))
"""
    )

    def exec(*cmd):
        process = do_run([sys.executable, "-m", "inmanta.app"] + list(cmd), cwd=snippetcompiler.project_dir)
        _, err = process.communicate(timeout=30)
        assert output in err.decode()

    no_cache_option = [] if cache_cf_files else ["--no-cache"]

    cl_compile = ["compile"] + no_cache_option
    cl_export = ["export", "-J", "out.json"] + no_cache_option

    exec(*cl_compile)
    exec(*cl_export)


def test_export_no_server(snippetcompiler):
    """
    Make sure that export to json works if no server is present.
    """
    snippetcompiler.setup_for_snippet("")
    process = do_run([sys.executable, "-m", "inmanta.app", "export", "-j", "out.json"], cwd=snippetcompiler.project_dir)
    stdout, stderr = process.communicate(timeout=30)
    assert process.returncode == 0, f"{stdout}\n\n{stderr}"


@pytest.mark.timeout(15)
@pytest.mark.parametrize_any(
    "cmd", [(["-X", "compile"]), (["compile", "-X"]), (["compile"]), (["export", "-X"]), (["-X", "export"]), (["export"])]
)
def test_minus_x_option(snippetcompiler, cmd):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    nuber attr
end
""",
        autostd=True,
    )

    process = do_run([sys.executable, "-m", "inmanta.app"] + cmd, cwd=snippetcompiler.project_dir)
    out, err = process.communicate(timeout=30)
    if "-X" in cmd:
        assert "inmanta.ast.TypeNotFoundException: could not find type nuber in namespace" in str(err)
    else:
        assert "inmanta.ast.TypeNotFoundException: could not find type nuber in namespace" not in str(err)


@pytest.mark.timeout(20)
def test_warning_config_dir_option_on_server_command(tmpdir):
    non_existing_dir = os.path.join(tmpdir, "non_existing_dir")
    assert not os.path.isdir(non_existing_dir)
    (args, _) = get_command(tmpdir, stdout_log_level=3, config_dir=non_existing_dir)
    (stdout, _, _) = run_without_tty(args, killtime=10, termtime=5)
    stdout = "".join(stdout)
    assert "Starting server endpoint" in stdout
    assert f"Config directory {non_existing_dir} doesn't exist" in stdout


@pytest.mark.timeout(20)
def test_warning_min_c_option_file_doesnt_exist(snippetcompiler, tmpdir):
    non_existing_config_file = os.path.join(tmpdir, "non_existing_config_file")
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    int attr
end
""",
        autostd=True,
    )
    config_options = ["-c", non_existing_config_file, "-vvv"]
    args = [sys.executable, "-m", "inmanta.app"] + config_options + ["compile"]
    process = do_run(args, cwd=snippetcompiler.project_dir)
    out, err = process.communicate(timeout=30)
    out = out.decode()
    err = err.decode()
    all_output = out + err

    assert process.returncode == 0, f"Error occured: {all_output}"

    assert "Starting compile" in all_output
    assert "Compile done" in all_output
    assert f"Config file {non_existing_config_file} doesn't exist" in all_output


@pytest.mark.parametrize_any(
    "with_tty, version_should_be_shown, regexes_required_lines, regexes_forbidden_lines",
    [
        (False, True, [r"Inmanta Service Orchestrator", r"Compiler version: ", r"Extensions:", r"\s*\* core:"], []),
        (True, True, [r"Inmanta Service Orchestrator", r"Compiler version: ", r"Extensions:", r"\s*\* core:"], []),
        (False, False, [], [r"Inmanta Service Orchestrator", r"Compiler version: ", r"Extensions:", r"\s*\* core:"]),
        (True, False, [], [r"Inmanta Service Orchestrator", r"Compiler version: ", r"Extensions:", r"\s*\* core:"]),
    ],
)
@pytest.mark.timeout(20)
def test_version_argument_is_set(tmpdir, with_tty, version_should_be_shown, regexes_required_lines, regexes_forbidden_lines):
    (args, log_dir) = get_command(tmpdir, version=version_should_be_shown)
    if with_tty:
        (stdout, _, _) = run_with_tty(args, killtime=15, termtime=10)
    else:
        (stdout, _, _) = run_without_tty(args, killtime=15, termtime=10)
    assert len(stdout) != 0
    check_logs(stdout, regexes_required_lines, regexes_forbidden_lines, False)


def test_init_project(tmpdir):
    args = [sys.executable, "-m", "inmanta.app", "project", "init", "-n", "test-project", "-o", tmpdir, "--default"]
    (stdout, stderr, return_code) = run_without_tty(args, killtime=15, termtime=10)
    test_project_path = os.path.join(tmpdir, "test-project")
    assert return_code == 0
    assert os.path.exists(test_project_path)
    (stdout, stderr, return_code) = run_without_tty(args, killtime=15, termtime=10)
    assert return_code != 0
    assert any("already exists" in error for error in stderr)


def test_compiler_summary_reporter(monkeypatch, capsys) -> None:
    """
    Test whether the CompileSummaryReporter class produces correct output.
    """
    # Success
    summary_reporter = CompileSummaryReporter()
    with summary_reporter.compiler_exception.capture():
        pass
    assert not summary_reporter.is_failure()
    summary_reporter.print_summary(show_stack_traces=True)
    assert re.match(r"\n=+ SUCCESS =+\n", capsys.readouterr().err)

    # Compile failure
    summary_reporter = CompileSummaryReporter()
    with summary_reporter.compiler_exception.capture():
        raise Exception("This is a compilation failure")
    assert summary_reporter.is_failure()
    summary_reporter.print_summary(show_stack_traces=False)
    output = capsys.readouterr().err
    assert re.match(r"\n=+ COMPILATION FAILURE =+\nError: This is a compilation failure\n", output)
    assert "= EXCEPTION TRACE =" not in output
    summary_reporter.print_summary(show_stack_traces=True)
    output = capsys.readouterr().err
    assert re.match(
        r"\n=+ EXCEPTION TRACE =+\n(.|\n)*\n=+ COMPILATION FAILURE =+\nError: This is a compilation failure\n", output
    )

    # Compile failure and export failure
    summary_reporter = CompileSummaryReporter()
    with summary_reporter.compiler_exception.capture():
        raise Exception("This is a compilation failure")
    with summary_reporter.exporter_exception.capture():
        raise Exception("This is an export failure")
    assert summary_reporter.is_failure()
    summary_reporter.print_summary(show_stack_traces=False)
    output = capsys.readouterr().err
    assert re.match(r"\n=+ COMPILATION FAILURE =+\nError: This is a compilation failure\n", output)
    assert "= EXCEPTION TRACE =" not in output

    # Export failure
    summary_reporter = CompileSummaryReporter()
    with summary_reporter.compiler_exception.capture():
        pass
    with summary_reporter.exporter_exception.capture():
        raise Exception("This is an export failure")
    assert summary_reporter.is_failure()
    summary_reporter.print_summary(show_stack_traces=False)
    output = capsys.readouterr().err
    assert re.match(r"\n=+ EXPORT FAILURE =+\nError: This is an export failure\n", output)
    assert "= EXCEPTION TRACE =" not in output
    summary_reporter.print_summary(show_stack_traces=True)
    output = capsys.readouterr().err
    assert re.match(r"\n=+ EXCEPTION TRACE =+\n(.|\n)*\n=+ EXPORT FAILURE =+\nError: This is an export failure\n", output)


def test_validate_logging_config(tmpdir, monkeypatch):
    """
    Test the validate-logging-config command.
    """
    _, stderr, returncode = run_without_tty(
        args=[sys.executable, "-m", "inmanta.app", "validate-logging-config"],
    )
    assert returncode == 1
    assert any("No logging configuration found." in line for line in stderr)

    logging_config_file = os.path.join(tmpdir, "non-existing-file")
    _, stderr, returncode = run_without_tty(
        args=[sys.executable, "-m", "inmanta.app", "--logging-config", logging_config_file, "validate-logging-config"],
    )
    assert returncode == 1
    assert any(f"Logging config file {logging_config_file} doesn't exist" in line for line in stderr)

    logging_config_file = os.path.join(tmpdir, "logging_config.yml")
    # Write logging config that contains a syntax error
    with open(logging_config_file, "w") as fh:
        fh.write(
            """
        formatters:
          server_log_formatter:
            format: '%(levelname)-8s %(name)-10s %(message)s'
        handlers   # HERE A COLON IS MISSING
          server_handler:
            class: logging.StreamHandler
            level: INFO
            formatter: server_log_formatter
            stream: ext://sys.stdout
        root:
          handlers:
          - server_handler
          level: DEBUG
        version: 1
        disable_existing_loggers: false
        """
        )
    _, stderr, returncode = run_without_tty(
        args=[sys.executable, "-m", "inmanta.app", "--logging-config", logging_config_file, "validate-logging-config"],
    )
    assert returncode == 1
    assert any("could not find expected ':'" in line for line in stderr)

    # Simple logging config that writes all logs to stdout
    with open(logging_config_file, "w") as fh:
        fh.write(
            """
        formatters:
          server_log_formatter:
            format: '%(levelname)-8s %(name)-10s %(message)s'
        handlers:
          server_handler:
            class: logging.StreamHandler
            level: INFO
            formatter: server_log_formatter
            stream: ext://sys.stdout
        root:
          handlers:
          - server_handler
          level: DEBUG
        version: 1
        disable_existing_loggers: false
        """
        )
    stdout, stderr, returncode = run_without_tty(
        args=[sys.executable, "-m", "inmanta.app", "--logging-config", logging_config_file, "validate-logging-config"],
    )
    assert returncode == 0
    assert any(f"Using logging config from file {logging_config_file}" in line for line in stderr)
    assert any(
        "Emitting log line 'Log line from Inmanta server at level <LEVEL>' using logger 'inmanta.protocol.rest.server'" in line
        for line in stderr
    )
    assert any("INFO     inmanta.protocol.rest.server Log line from Inmanta server at level INFO" in line for line in stdout)
    assert not any(
        "DEBUG    inmanta.protocol.rest.server Log line from Inmanta server at level DEBUG" in line for line in stdout
    )

    # Write a logging config that works.
    #  * Write all logs to server.log
    #  * Write resource-action-log to different file.
    log_dir = os.path.join(tmpdir, "log")
    os.mkdir(log_dir)
    assert not os.listdir(log_dir)
    logging_config_file = os.path.join(tmpdir, "logging_config.yml.tmpl")
    with open(logging_config_file, "w") as fh:
        fh.write(
            f"""
        formatters:
          server_log_formatter:
            format: '%(levelname)-8s %(name)-10s %(message)s'
          resource_action_log_formatter:
            format: '%(name)-10s %(levelname)-8s %(message)s'
        handlers:
          server_handler:
            class: logging.handlers.WatchedFileHandler
            filename: {log_dir}/server.log
            formatter: server_log_formatter
            level: WARNING
            mode: a+
          scheduler_resource_action_handler:
            class: logging.handlers.WatchedFileHandler
            filename: {log_dir}/resource-action-{{environment}}.log
            formatter: resource_action_log_formatter
            level: ERROR
        loggers:
          inmanta.resource_action:
            handlers:
            - scheduler_resource_action_handler
            level: DEBUG
            propagate: true
        root:
          handlers:
          - server_handler
          level: DEBUG
        version: 1
        disable_existing_loggers: false
        """
        )
    env_id = uuid.uuid4()
    stdout, stderr, returncode = run_without_tty(
        args=[
            sys.executable,
            "-m",
            "inmanta.app",
            "--logging-config",
            logging_config_file,
            "validate-logging-config",
            "-e",
            str(env_id),
        ],
    )
    assert any(f"Using logging config from file {logging_config_file}" in line for line in stderr)
    assert any(
        "Emitting log line 'Log line from callback at level <LEVEL>' using logger 'inmanta_lsm.callback'" in line
        for line in stderr
    )
    assert any(
        f"Emitting log line 'Log line for resource action log at level <LEVEL>' using logger 'inmanta.resource_action.{env_id}'"
        in line
        for line in stderr
    )
    assert returncode == 0
    server_log_file = os.path.join(log_dir, "server.log")
    resource_action_log_file = os.path.join(log_dir, f"resource-action-{env_id}.log")
    files_in_log_dir = os.listdir(log_dir)
    assert sorted(files_in_log_dir) == sorted([os.path.basename(server_log_file), os.path.basename(resource_action_log_file)])

    with open(server_log_file, "r") as fh:
        content_log_file = fh.read()
    assert "WARNING  inmanta_lsm.callback Log line from callback at level WARNING" in content_log_file
    assert "INFO     inmanta_lsm.callback Log line from callback at level INFO" not in content_log_file

    with open(resource_action_log_file, "r") as fh:
        content_log_file = fh.read()
    assert f"inmanta.resource_action.{env_id} ERROR    Log line for resource action log at level ERROR" in content_log_file
    assert f"inmanta.resource_action.{env_id} INFO     Log line for resource action log at level INFO" not in content_log_file
    assert "WARNING  inmanta_lsm.callback Log line from callback at level WARNING" not in content_log_file
    assert "INFO     inmanta_lsm.callback Log line from callback at level INFO" not in content_log_file

    # Test whether component config file is picked up correctly
    monkeypatch.chdir(tmpdir)
    logging_config_file = os.path.join(tmpdir, f"{logging_config_file}.yml")
    dot_inmanta_file = os.path.join(tmpdir, ".inmanta")
    for component_name in ["server", "scheduler", "compiler"]:
        with open(logging_config_file, "w") as fh:
            fh.write(
                f"""
                    formatters:
                      server_log_formatter:
                        format: '{component_name} -- %(message)s'
                    handlers:
                      server_handler:
                        class: logging.StreamHandler
                        level: INFO
                        formatter: server_log_formatter
                        stream: ext://sys.stdout
                    root:
                      handlers:
                      - server_handler
                      level: DEBUG
                    version: 1
                    disable_existing_loggers: false
                """
            )
        with open(dot_inmanta_file, "w") as fh:
            fh.write(
                f"""
                    [logging]
                    {component_name} = {logging_config_file}
                """
            )
        stdout, stderr, returncode = run_without_tty(
            args=[sys.executable, "-m", "inmanta.app", "validate-logging-config", component_name],
        )
        assert any(f"Using logging config from file {logging_config_file}" in line for line in stderr)
        assert any(
            "Emitting log line 'Log line from Inmanta server at level <LEVEL>' using logger 'inmanta.protocol.rest.server'"
            in line
            for line in stderr
        )
        assert any(f"{component_name} -- Log line from Inmanta server at level INFO" in line for line in stdout)
        assert returncode == 0



"""
Copyright 2022 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

"""
Tests to verify correctness/compatibility of code snippets in the docs.
"""

import os
import textwrap
import uuid
from collections import abc, defaultdict

import py
import pytest

import inmanta.types
from inmanta import compiler, data
from utils import v1_module_from_template, wait_until_deployment_finishes

DOCS_DIR: str = os.path.join(os.path.dirname(__file__), "..", "docs")


@pytest.mark.slowtest
async def test_docs_snippets_language_unknowns(snippetcompiler, capsys) -> None:
    """
    Test the code snippets that explain how unknowns flow through the model.
    """
    with open(os.path.join(os.path.dirname(__file__), "../docs/language/unknowns_simple.cf")) as fh:
        simple_model: str = fh.read()

    with open(os.path.join(os.path.dirname(__file__), "../docs/language/unknowns_multi.cf")) as fh:
        list_model: str = fh.read()

    # TODO: convert_unknowns typing port
    model = textwrap.dedent(
        f"""\
        {simple_model}

        import tests

        assert = true
        assert = std::is_unknown(a)
        assert = [1, 2, "?", 3] == tests::convert_unknowns(b, "?")
        assert = std::is_unknown(c)
        assert = d
        assert = std::is_unknown(e)
        assert = std::is_unknown(f)
        assert = std::is_unknown(g)
        assert = [1, 2, "?"] == [x == "?" ? x : x.n for x in tests::convert_unknowns(h, "?")]
        assert = [1, 2] == [x.n for x in i]
        """
    )
    snippetcompiler.setup_for_snippet(model, ministd=True)
    compiler.do_compile()

    out, _ = capsys.readouterr()
    assert out == textwrap.dedent(
        """\
        This message is printed twice! x=1
        This message is printed twice! x=2
        """
    )

    model = textwrap.dedent(
        f"""\
        {list_model}

        import tests

        assert = true
        assert = ["?", 3, "?", 5] == tests::convert_unknowns(a, "?")
        assert = std::is_unknown(b)
        assert = ["?", "?"] == tests::convert_unknowns(c, "?")
        assert = std::is_unknown(d)
        """
    )
    snippetcompiler.setup_for_snippet(model, ministd=True)
    compiler.do_compile()


@pytest.mark.slowtest
async def test_docs_snippet_partial_compile(
    tmpdir: py.path.local, snippetcompiler, modules_dir: str, server, client, environment: str
) -> None:
    """
    Verify that the partial compile example model is valid and is in fact equivalent to the full model it is compared to.
    """
    env_id: uuid.UUID = uuid.UUID(environment)
    snippets_dir: str = os.path.join(DOCS_DIR, "model_developers", "resource_sets")
    version: int

    handler_module_name: str = "host_handlers"
    v1_module_from_template(
        os.path.join(modules_dir, "minimalv1module"),
        str(tmpdir.join(handler_module_name)),
        new_name=handler_module_name,
        new_content_init_py=textwrap.dedent(
            """
            from inmanta.execute.proxy import DynamicProxy
            from inmanta.export import Exporter
            from inmanta.resources import Resource, resource

            @resource("__config__::Host", agent="agent.agentname", id_attribute="full_id")
            class Host(Resource):
                fields = ("network_id", "host_id", "full_id",)

                def get_network_id(exporter: Exporter, obj: DynamicProxy) -> int:
                    return obj.network.id

                def get_host_id(exporter: Exporter, obj: DynamicProxy) -> int:
                    return obj.id

                def get_full_id(exporter: Exporter, obj: DynamicProxy) -> tuple[int, str]:
                    return (Host.get_network_id(exporter, obj), Host.get_host_id(exporter, obj))
            """.strip(
                "\n"
            )
        ),
    )

    def setup_model(base: str) -> None:
        """
        Add handlers for the base model and set up the snippetcompiler.
        """
        handlers_addition: str = (
            f"""
            import {handler_module_name} as handler

            # add dummy agent attribute for the handler
            Host.agent [1] -- std::AgentConfig
            implementation bind_agent for Host:
                self.agent = std::AgentConfig[agentname="host_agent"]
            end
            implement Host using bind_agent
        """.strip()
        )
        full_model: str = "\n".join((base, handlers_addition))
        snippetcompiler.setup_for_snippet(full_model, add_to_module_path=[str(tmpdir)], autostd=True)

    async def get_hosts_by_network(version: int) -> abc.Mapping[int, abc.Set[int]]:
        resources: abc.Sequence[data.Resource] = await data.Resource.get_resources_for_version(env_id, version)
        hosts_by_network: dict[int, set[int]] = defaultdict(set)
        for resource in resources:
            if resource.resource_type == inmanta.types.ResourceType("__config__::Host"):
                hosts_by_network[resource.attributes["network_id"]].add(resource.attributes["host_id"])
        return hosts_by_network

    # initial export
    with open(os.path.join(snippets_dir, "basic_example_full.cf")) as fd:
        setup_model(fd.read())
    version, _ = await snippetcompiler.do_export_and_deploy()
    hosts_by_network_full: abc.Mapping[int, abc.Set[int]] = await get_hosts_by_network(version)
    assert len(hosts_by_network_full) == 1000
    for network in hosts_by_network_full:
        assert len(hosts_by_network_full[network]) == 5

    # partial export: verify that only the example's set has changed
    with open(os.path.join(snippets_dir, "basic_example_partial.cf")) as fd:
        setup_model(fd.read())
    version, _ = await snippetcompiler.do_export_and_deploy(partial_compile=True)
    hosts_by_network_partial: abc.Mapping[int, abc.Set[int]] = await get_hosts_by_network(version)
    assert len(hosts_by_network_partial) == 1000
    for network in hosts_by_network_partial:
        assert len(hosts_by_network_partial[network]) == (1 if network == 0 else 5)

    # full equivalent export: verify that it is indeed equivalent
    with open(os.path.join(snippets_dir, "basic_example_full_result.cf")) as fd:
        setup_model(fd.read())
    version, _ = await snippetcompiler.do_export_and_deploy()
    hosts_by_network_equivalent: abc.Mapping[int, abc.Set[int]] = await get_hosts_by_network(version)
    assert hosts_by_network_equivalent == hosts_by_network_partial


@pytest.mark.slowtest
async def test_docs_snippets_unmanaged_resources_basic(
    tmpdir: py.path.local, snippetcompiler, modules_dir: str, server, client, environment: str, agent, clienthelper
) -> None:
    """
    Test the basic_example code snippets used in the documentation to explain the usage of unmanaged_resources.
    """
    result = await client.set_setting(environment, data.AUTO_DEPLOY, True)
    assert result.code == 200

    cf_file_path = os.path.join(os.path.dirname(__file__), "../docs/model_developers/unmanaged_resources/basic_example.cf")
    with open(cf_file_path) as fh:
        cf_file_content = fh.read()

    attributes_discovered_resources = [{"host": "localhost", "interface_name": "eth0", "ip_address": "10.10.10.10"}]
    init_py_path = os.path.join(
        os.path.dirname(__file__), "../docs/model_developers/unmanaged_resources/basic_example_handler.py"
    )
    with open(init_py_path) as fh:
        init_py_content = fh.read()
    init_py_content = init_py_content.replace("raise NotImplementedError()", f"return {attributes_discovered_resources}")

    module_name = "my_module"
    v1_module_from_template(
        source_dir=os.path.join(modules_dir, "minimalv1module"),
        dest_dir=str(tmpdir.join(module_name)),
        new_name=module_name,
        new_content_init_cf=cf_file_content,
        new_content_init_py=init_py_content,
    )

    model = textwrap.dedent(
        """
        import my_module

        host = std::Host(ip="127.0.0.1", name="agent1", os=std::linux)
        my_module::InterfaceDiscovery(host=host)
        """
    )
    snippetcompiler.setup_for_snippet(model, add_to_module_path=[str(tmpdir)], use_pip_config_file=True, autostd=True)
    version, _ = await snippetcompiler.do_export_and_deploy()

    await clienthelper.wait_for_released(version)
    await wait_until_deployment_finishes(client, environment, version=version)

    result = await client.discovered_resources_get_batch(tid=environment)
    assert result.code == 200
    assert len(result.result["data"]) == 1
    assert result.result["data"][0]["discovered_resource_id"] == "my_module::Interface[localhost,interface_name=eth0]"
    assert result.result["data"][0]["values"] == attributes_discovered_resources[0]


@pytest.mark.slowtest
async def test_docs_snippets_unmanaged_resources_shared_attributes(
    tmpdir: py.path.local, snippetcompiler, modules_dir: str, server, client, environment: str, agent, clienthelper
) -> None:
    """
    Test the shared_attributes_example code snippets used in the documentation to explain the usage of unmanaged_resources.
    """
    result = await client.set_setting(environment, data.AUTO_DEPLOY, True)
    assert result.code == 200

    cf_file_path = os.path.join(
        os.path.dirname(__file__), "../docs/model_developers/unmanaged_resources/shared_attributes_example.cf"
    )
    with open(cf_file_path) as fh:
        cf_file_content = fh.read()

    attributes_discovered_resources = [
        {"host": "localhost", "interface_name": "eth0", "ip_address": "10.10.10.10"},
        {"host": "localhost", "interface_name": "eth1", "ip_address": "20.20.20.20"},
    ]
    init_py_path = os.path.join(
        os.path.dirname(__file__), "../docs/model_developers/unmanaged_resources/shared_attributes_example_handler.py"
    )
    with open(init_py_path) as fh:
        init_py_content = fh.read()
    init_py_content = init_py_content.replace("raise NotImplementedError()", "pass", 6)
    init_py_content = init_py_content.replace("raise NotImplementedError()", f"return {attributes_discovered_resources}")

    module_name = "my_module"
    v1_module_from_template(
        source_dir=os.path.join(modules_dir, "minimalv1module"),
        dest_dir=str(tmpdir.join(module_name)),
        new_name=module_name,
        new_content_init_cf=cf_file_content,
        new_content_init_py=init_py_content,
    )

    model = textwrap.dedent(
        """
        import my_module

        host = std::Host(ip="127.0.0.1", name="agent1", os=std::linux)
        credentials = my_module::Credentials(username="test", password="test")
        my_module::InterfaceDiscovery(name_filter="eth[1-9]", host=host, credentials=credentials)
        """
    )
    snippetcompiler.setup_for_snippet(model, add_to_module_path=[str(tmpdir)], use_pip_config_file=True, autostd=True)
    version, _ = await snippetcompiler.do_export_and_deploy()

    await clienthelper.wait_for_released(version)

    await wait_until_deployment_finishes(client, environment, version=version)

    result = await client.discovered_resources_get_batch(tid=environment)
    assert result.code == 200
    assert len(result.result["data"]) == 1
    assert result.result["data"][0]["discovered_resource_id"] == "my_module::Interface[localhost,interface_name=eth1]"
    assert result.result["data"][0]["values"] == attributes_discovered_resources[1]



"""
Copyright 2021 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import os

import inmanta.util
import packaging.utils
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap


class PreservativeYamlParser:
    """
    A Yaml parser that tries to preserve comments and the order of elements.
    This parser doesn't preserve indentation.
    """

    @classmethod
    def _get_parser(cls) -> YAML:
        parser = YAML()
        # Make sure the indentation settings are used consistently
        parser.indent(mapping=2, sequence=4, offset=2)
        return parser

    @classmethod
    def parse(cls, filename: str) -> CommentedMap:
        parser = cls._get_parser()
        with open(filename, encoding="utf-8") as fd:
            return parser.load(fd)

    @classmethod
    def dump(cls, filename: str, content: CommentedMap) -> None:
        parser = cls._get_parser()
        with open(filename, "w", encoding="utf-8") as fd:
            parser.dump(content, stream=fd)


class RequirementsTxtParser:
    """
    Parser for a requirements.txt file
    """

    @classmethod
    def parse(cls, filename: str) -> list[inmanta.util.CanonicalRequirement]:
        """
        Get all the requirements in `filename` as a list of `Requirement` instances.
        """
        return [inmanta.util.parse_requirement(requirement=r) for r in cls.parse_requirements_as_strs(filename)]

    @classmethod
    def parse_requirements_as_strs(cls, filename: str) -> list[str]:
        """
        Get all the requirements in `filename` as a list of strings.
        """
        if os.path.exists(filename):
            with open(filename, encoding="utf-8") as fd:
                requirements_txt_content = fd.read()
                req_lines = [x.strip() for x in requirements_txt_content.split("\n") if len(x.strip()) > 0]
                req_lines = cls._remove_comments(req_lines)
                req_lines = cls._remove_line_continuations(req_lines)
                return list(req_lines)
        else:
            return []

    @classmethod
    def get_content_with_dep_removed(cls, filename: str, remove_dep_on_pkg: str) -> str:
        """
        Returns the content of the requirements.txt file with the dependency on `remove_dep_on_pkg` removed.
        This method preserves all the comments.
        """
        if not os.path.exists(filename):
            raise Exception(f"File {filename} doesn't exist")

        removed_dependency = packaging.utils.canonicalize_name(remove_dep_on_pkg)
        result = ""
        line_continuation_buffer = ""
        with open(filename, encoding="utf-8") as fd:
            for line in fd.readlines():
                if line_continuation_buffer:
                    line_continuation_buffer += line
                    if not line.endswith("\\"):
                        if (
                            inmanta.util.parse_requirement(
                                requirement=inmanta.util.remove_comment_part_from_specifier(line_continuation_buffer)
                            ).name
                            != removed_dependency
                        ):
                            result += line_continuation_buffer
                        line_continuation_buffer = ""
                elif not line.strip() or line.strip().startswith("#"):
                    result += line
                elif line.endswith("\\"):
                    line_continuation_buffer = line
                elif (
                    inmanta.util.parse_requirement(requirement=inmanta.util.remove_comment_part_from_specifier(line)).name
                    != removed_dependency
                ):
                    result += line
                else:
                    # Dependency matches `remove_dep_on_pkg` => Remove line from result
                    pass
        return result

    @classmethod
    def _remove_comments(cls, lines: list[str]) -> list[str]:
        """
        This method removes elements from the given list that only include comments. If the element
        combines a comment with a version constraint, the comment part is removed from the element.

        :param lines: The lines from a requirements.txt file with all empty lines removes.
        """
        result = []
        for line in lines:
            if line.strip().startswith("#"):
                continue
            if " #" in line:
                line_without_comment = line.split(" #", maxsplit=1)[0]
                result.append(line_without_comment)
            else:
                result.append(line)
        return result

    @classmethod
    def _remove_line_continuations(cls, lines: list[str]) -> list[str]:
        """
        Join two different list elements together if they are separated by a line continuation token.

        :param lines: The lines from a requirements.txt file with all empty lines removes.
        """
        result = []
        line_continuation_buffer = ""
        for line in lines:
            if line.endswith("\\"):
                line_continuation_buffer = f"{line_continuation_buffer}{line[0:-1]}"
            else:
                if line_continuation_buffer:
                    result.append(f"{line_continuation_buffer}{line}")
                    line_continuation_buffer = ""
                else:
                    result.append(line)
        return result
